from typing import Any
from ...arctrl_file_system.path import combine_many
from ...fable_library.guid import new_guid
from ...fable_library.reg_exp import (get_item, groups)
from ...fable_library.string_ import starts_with_exact
from .regex import ActivePatterns__007CRegex_007C__007C

def try_check_valid_characters(identifier: str) -> bool:
    if ActivePatterns__007CRegex_007C__007C("^[a-zA-Z0-9_\\- ]+$", identifier) is not None:
        return True

    else: 
        return False



def check_valid_characters(identifier: str) -> None:
    if try_check_valid_characters(identifier):
        pass

    else: 
        raise Exception(("New identifier \"" + identifier) + "\" contains forbidden characters! Allowed characters are: letters, digits, underscore (_), dash (-) and whitespace ( ).")



def create_missing_identifier(__unit: None=None) -> str:
    def _arrow695(__unit: None=None) -> str:
        copy_of_struct: str = new_guid()
        return str(copy_of_struct)

    return "MISSING_IDENTIFIER_" + _arrow695()


def is_missing_identifier(str_1: str) -> bool:
    return starts_with_exact(str_1, "MISSING_IDENTIFIER_")


def remove_missing_identifier(str_1: str) -> str:
    if starts_with_exact(str_1, "MISSING_IDENTIFIER_"):
        return ""

    else: 
        return str_1



def Assay_identifierFromFileName(file_name: str) -> str:
    active_pattern_result: Any | None = ActivePatterns__007CRegex_007C__007C("^(assays(\\/|\\\\))?(?<identifier>[a-zA-Z0-9_\\- ]+)((\\/|\\\\)isa.assay.xlsx)?$", file_name)
    if active_pattern_result is not None:
        m: Any = active_pattern_result
        return get_item(groups(m), "identifier") or ""

    else: 
        raise Exception(("Cannot parse assay identifier from FileName `" + file_name) + "`")



def Assay_tryIdentifierFromFileName(file_name: str) -> str | None:
    active_pattern_result: Any | None = ActivePatterns__007CRegex_007C__007C("^(assays(\\/|\\\\))?(?<identifier>[a-zA-Z0-9_\\- ]+)((\\/|\\\\)isa.assay.xlsx)?$", file_name)
    if active_pattern_result is not None:
        m: Any = active_pattern_result
        return get_item(groups(m), "identifier") or ""

    else: 
        return None



def Assay_fileNameFromIdentifier(identifier: str) -> str:
    check_valid_characters(identifier)
    return combine_many(["assays", identifier, "isa.assay.xlsx"])


def Assay_tryFileNameFromIdentifier(identifier: str) -> str | None:
    if try_check_valid_characters(identifier):
        return combine_many(["assays", identifier, "isa.assay.xlsx"])

    else: 
        return None



def Assay_datamapFileNameFromIdentifier(identifier: str) -> str:
    check_valid_characters(identifier)
    return combine_many(["assays", identifier, "isa.datamap.xlsx"])


def Assay_tryDatamapFileNameFromIdentifier(identifier: str) -> str | None:
    if try_check_valid_characters(identifier):
        return combine_many(["assays", identifier, "isa.datamap.xlsx"])

    else: 
        return None



def Study_identifierFromFileName(file_name: str) -> str:
    active_pattern_result: Any | None = ActivePatterns__007CRegex_007C__007C("^(studies(\\/|\\\\))?(?<identifier>[a-zA-Z0-9_\\- ]+)((\\/|\\\\)isa.study.xlsx)?$", file_name)
    if active_pattern_result is not None:
        m: Any = active_pattern_result
        return get_item(groups(m), "identifier") or ""

    else: 
        raise Exception(("Cannot parse study identifier from FileName `" + file_name) + "`")



def Study_tryIdentifierFromFileName(file_name: str) -> str | None:
    active_pattern_result: Any | None = ActivePatterns__007CRegex_007C__007C("^(studies(\\/|\\\\))?(?<identifier>[a-zA-Z0-9_\\- ]+)((\\/|\\\\)isa.study.xlsx)?$", file_name)
    if active_pattern_result is not None:
        m: Any = active_pattern_result
        return get_item(groups(m), "identifier") or ""

    else: 
        return None



def Study_fileNameFromIdentifier(identifier: str) -> str:
    check_valid_characters(identifier)
    return combine_many(["studies", identifier, "isa.study.xlsx"])


def Study_tryFileNameFromIdentifier(identifier: str) -> str | None:
    if try_check_valid_characters(identifier):
        return combine_many(["studies", identifier, "isa.study.xlsx"])

    else: 
        return None



def Study_datamapFileNameFromIdentifier(identifier: str) -> str:
    check_valid_characters(identifier)
    return combine_many(["studies", identifier, "isa.datamap.xlsx"])


def Study_tryDatamapFileNameFromIdentifier(identifier: str) -> str | None:
    if try_check_valid_characters(identifier):
        return combine_many(["studies", identifier, "isa.datamap.xlsx"])

    else: 
        return None



def Workflow_identifierFromFileName(file_name: str) -> str:
    active_pattern_result: Any | None = ActivePatterns__007CRegex_007C__007C("^(workflows(\\/|\\\\))?(?<identifier>[a-zA-Z0-9_\\- ]+)((\\/|\\\\)isa.workflow.xlsx)?$", file_name)
    if active_pattern_result is not None:
        m: Any = active_pattern_result
        return get_item(groups(m), "identifier") or ""

    else: 
        raise Exception(("Cannot parse workflow identifier from FileName `" + file_name) + "`")



def Workflow_tryIdentifierFromFileName(file_name: str) -> str | None:
    active_pattern_result: Any | None = ActivePatterns__007CRegex_007C__007C("^(workflows(\\/|\\\\))?(?<identifier>[a-zA-Z0-9_\\- ]+)((\\/|\\\\)isa.workflow.xlsx)?$", file_name)
    if active_pattern_result is not None:
        m: Any = active_pattern_result
        return get_item(groups(m), "identifier") or ""

    else: 
        return None



def Workflow_fileNameFromIdentifier(identifier: str) -> str:
    check_valid_characters(identifier)
    return combine_many(["workflows", identifier, "isa.workflow.xlsx"])


def Workflow_tryFileNameFromIdentifier(identifier: str) -> str | None:
    if try_check_valid_characters(identifier):
        return combine_many(["workflows", identifier, "isa.workflow.xlsx"])

    else: 
        return None



def Workflow_cwlFileNameFromIdentifier(identifier: str) -> str:
    check_valid_characters(identifier)
    return combine_many(["workflows", identifier, "workflow.cwl"])


def Workflow_tryCwlFileNameFromIdentifier(identifier: str) -> str | None:
    if try_check_valid_characters(identifier):
        return combine_many(["workflows", identifier, "workflow.cwl"])

    else: 
        return None



def Workflow_datamapFileNameFromIdentifier(identifier: str) -> str:
    check_valid_characters(identifier)
    return combine_many(["workflows", identifier, "isa.datamap.xlsx"])


def Workflow_tryDatamapFileNameFromIdentifier(identifier: str) -> str | None:
    if try_check_valid_characters(identifier):
        return combine_many(["workflows", identifier, "isa.datamap.xlsx"])

    else: 
        return None



def Run_identifierFromFileName(file_name: str) -> str:
    active_pattern_result: Any | None = ActivePatterns__007CRegex_007C__007C("^(runs(\\/|\\\\))?(?<identifier>[a-zA-Z0-9_\\- ]+)((\\/|\\\\)isa.run.xlsx)?$", file_name)
    if active_pattern_result is not None:
        m: Any = active_pattern_result
        return get_item(groups(m), "identifier") or ""

    else: 
        raise Exception(("Cannot parse run identifier from FileName `" + file_name) + "`")



def Run_tryIdentifierFromFileName(file_name: str) -> str | None:
    active_pattern_result: Any | None = ActivePatterns__007CRegex_007C__007C("^(runs(\\/|\\\\))?(?<identifier>[a-zA-Z0-9_\\- ]+)((\\/|\\\\)isa.run.xlsx)?$", file_name)
    if active_pattern_result is not None:
        m: Any = active_pattern_result
        return get_item(groups(m), "identifier") or ""

    else: 
        return None



def Run_fileNameFromIdentifier(identifier: str) -> str:
    check_valid_characters(identifier)
    return combine_many(["runs", identifier, "isa.run.xlsx"])


def Run_tryFileNameFromIdentifier(identifier: str) -> str | None:
    if try_check_valid_characters(identifier):
        return combine_many(["runs", identifier, "isa.run.xlsx"])

    else: 
        return None



def Run_cwlFileNameFromIdentifier(identifier: str) -> str:
    check_valid_characters(identifier)
    return combine_many(["runs", identifier, "run.cwl"])


def Run_tryCwlFileNameFromIdentifier(identifier: str) -> str | None:
    if try_check_valid_characters(identifier):
        return combine_many(["runs", identifier, "run.cwl"])

    else: 
        return None



def Run_ymlFileNameFromIdentifier(identifier: str) -> str:
    check_valid_characters(identifier)
    return combine_many(["runs", identifier, "run.yml"])


def Run_tryYmlFileNameFromIdentifier(identifier: str) -> str | None:
    if try_check_valid_characters(identifier):
        return combine_many(["runs", identifier, "run.yml"])

    else: 
        return None



def Run_datamapFileNameFromIdentifier(identifier: str) -> str:
    check_valid_characters(identifier)
    return combine_many(["runs", identifier, "isa.datamap.xlsx"])


def Run_tryDatamapFileNameFromIdentifier(identifier: str) -> str | None:
    if try_check_valid_characters(identifier):
        return combine_many(["runs", identifier, "isa.datamap.xlsx"])

    else: 
        return None



__all__ = ["try_check_valid_characters", "check_valid_characters", "create_missing_identifier", "is_missing_identifier", "remove_missing_identifier", "Assay_identifierFromFileName", "Assay_tryIdentifierFromFileName", "Assay_fileNameFromIdentifier", "Assay_tryFileNameFromIdentifier", "Assay_datamapFileNameFromIdentifier", "Assay_tryDatamapFileNameFromIdentifier", "Study_identifierFromFileName", "Study_tryIdentifierFromFileName", "Study_fileNameFromIdentifier", "Study_tryFileNameFromIdentifier", "Study_datamapFileNameFromIdentifier", "Study_tryDatamapFileNameFromIdentifier", "Workflow_identifierFromFileName", "Workflow_tryIdentifierFromFileName", "Workflow_fileNameFromIdentifier", "Workflow_tryFileNameFromIdentifier", "Workflow_cwlFileNameFromIdentifier", "Workflow_tryCwlFileNameFromIdentifier", "Workflow_datamapFileNameFromIdentifier", "Workflow_tryDatamapFileNameFromIdentifier", "Run_identifierFromFileName", "Run_tryIdentifierFromFileName", "Run_fileNameFromIdentifier", "Run_tryFileNameFromIdentifier", "Run_cwlFileNameFromIdentifier", "Run_tryCwlFileNameFromIdentifier", "Run_ymlFileNameFromIdentifier", "Run_tryYmlFileNameFromIdentifier", "Run_datamapFileNameFromIdentifier", "Run_tryDatamapFileNameFromIdentifier"]




from __future__ import annotations
from collections.abc import Callable
from typing import (Any, TypeVar)
from ...arctrl_core.comment import Comment
from ...arctrl_core.ontology_annotation import OntologyAnnotation
from ...arctrl_core.Process.factor import Factor
from ...fable_library.list import (choose, of_array, FSharpList)
from ...fable_library.option import (map, default_arg)
from ...fable_library.seq import map as map_1
from ...fable_library.types import Array
from ...fable_library.util import IEnumerable_1
from ...thoth_json_core.decode import (object, IOptionalGetter, string, resize_array, IGetters)
from ...thoth_json_core.types import (IEncodable, IEncoderHelpers_1, Decoder_1)
from ..comment import (ISAJson_encoder, ISAJson_decoder)
from ..encode import (try_include, try_include_seq)
from ..idtable import encode
from ..ontology_annotation import (OntologyAnnotation_ROCrate_genID, OntologyAnnotation_ISAJson_encoder, OntologyAnnotation_ISAJson_decoder)

__A_ = TypeVar("__A_")

def gen_id(f: Factor) -> str:
    match_value: str | None = f.Name
    if match_value is None:
        match_value_1: OntologyAnnotation | None = f.FactorType
        if match_value_1 is None:
            return "#EmptyFactor"

        else: 
            return ("#Factor/" + OntologyAnnotation_ROCrate_genID(match_value_1)) + ""


    else: 
        return ("#Factor/" + match_value) + ""



def encoder(id_map: Any | None, value: Factor) -> IEncodable:
    def f_1(value_1: Factor, id_map: Any=id_map, value: Any=value) -> IEncodable:
        def chooser(tupled_arg: tuple[str, IEncodable | None], value_1: Any=value_1) -> tuple[str, IEncodable] | None:
            def mapping(v_1: IEncodable, tupled_arg: Any=tupled_arg) -> tuple[str, IEncodable]:
                return (tupled_arg[0], v_1)

            return map(mapping, tupled_arg[1])

        def _arrow2330(value_2: str, value_1: Any=value_1) -> IEncodable:
            class ObjectExpr2329(IEncodable):
                def Encode(self, helpers: IEncoderHelpers_1[Any]) -> Any:
                    return helpers.encode_string(value_2)

            return ObjectExpr2329()

        def _arrow2332(value_4: str, value_1: Any=value_1) -> IEncodable:
            class ObjectExpr2331(IEncodable):
                def Encode(self, helpers_1: IEncoderHelpers_1[Any]) -> Any:
                    return helpers_1.encode_string(value_4)

            return ObjectExpr2331()

        def _arrow2333(oa: OntologyAnnotation, value_1: Any=value_1) -> IEncodable:
            return OntologyAnnotation_ISAJson_encoder(id_map, oa)

        def _arrow2334(comment: Comment, value_1: Any=value_1) -> IEncodable:
            return ISAJson_encoder(id_map, comment)

        values: FSharpList[tuple[str, IEncodable]] = choose(chooser, of_array([try_include("@id", _arrow2330, gen_id(value_1)), try_include("factorName", _arrow2332, value_1.Name), try_include("factorType", _arrow2333, value_1.FactorType), try_include_seq("comments", _arrow2334, default_arg(value_1.Comments, []))]))
        class ObjectExpr2337(IEncodable):
            def Encode(self, helpers_2: IEncoderHelpers_1[Any], value_1: Any=value_1) -> Any:
                def mapping_1(tupled_arg_1: tuple[str, IEncodable]) -> tuple[str, __A_]:
                    return (tupled_arg_1[0], tupled_arg_1[1].Encode(helpers_2))

                arg: IEnumerable_1[tuple[str, __A_]] = map_1(mapping_1, values)
                return helpers_2.encode_object(arg)

        return ObjectExpr2337()

    if id_map is not None:
        def _arrow2340(f_2: Factor, id_map: Any=id_map, value: Any=value) -> str:
            return gen_id(f_2)

        return encode(_arrow2340, f_1, value, id_map)

    else: 
        return f_1(value)



def _arrow2346(get: IGetters) -> Factor:
    def _arrow2342(__unit: None=None) -> str | None:
        object_arg: IOptionalGetter = get.Optional
        return object_arg.Field("factorName", string)

    def _arrow2343(__unit: None=None) -> OntologyAnnotation | None:
        object_arg_1: IOptionalGetter = get.Optional
        return object_arg_1.Field("factorType", OntologyAnnotation_ISAJson_decoder)

    def _arrow2345(__unit: None=None) -> Array[Comment] | None:
        arg_5: Decoder_1[Array[Comment]] = resize_array(ISAJson_decoder)
        object_arg_2: IOptionalGetter = get.Optional
        return object_arg_2.Field("comments", arg_5)

    return Factor(_arrow2342(), _arrow2343(), _arrow2345())


decoder: Decoder_1[Factor] = object(_arrow2346)

__all__ = ["gen_id", "encoder", "decoder"]




from __future__ import annotations
from typing import (Any, TypeVar)
from ...fable_library.option import default_arg
from ...fable_library.seq import (map, empty)
from ...fable_library.util import (to_enumerable, IEnumerable_1)
from ...thoth_json_core.decode import (object, IOptionalGetter, int_1, seq as seq_1, IGetters)
from ...thoth_json_core.encode import seq
from ...thoth_json_core.types import (IEncodable, IEncoderHelpers_1, Decoder_1)
from ..Cells.fs_cell import FsCell
from ..fs_column import FsColumn
from .cell import (encode_cols, encode_no_number, decode_cols)

__A_ = TypeVar("__A_")

def encode(col: FsColumn) -> IEncodable:
    def _arrow349(__unit: None=None, col: Any=col) -> IEncodable:
        value: int = col.Index or 0
        class ObjectExpr348(IEncodable):
            def Encode(self, helpers: IEncoderHelpers_1[Any]) -> Any:
                return helpers.encode_signed_integral_number(value)

        return ObjectExpr348()

    def mapping(cell: FsCell, col: Any=col) -> IEncodable:
        return encode_cols(cell)

    values: IEnumerable_1[tuple[str, IEncodable]] = to_enumerable([("number", _arrow349()), ("cells", seq(map(mapping, col.Cells)))])
    class ObjectExpr350(IEncodable):
        def Encode(self, helpers_1: IEncoderHelpers_1[Any], col: Any=col) -> Any:
            def mapping_1(tupled_arg: tuple[str, IEncodable]) -> tuple[str, __A_]:
                return (tupled_arg[0], tupled_arg[1].Encode(helpers_1))

            arg: IEnumerable_1[tuple[str, __A_]] = map(mapping_1, values)
            return helpers_1.encode_object(arg)

    return ObjectExpr350()


def encode_no_numbers(col: IEnumerable_1[FsCell]) -> IEncodable:
    def mapping(cell: FsCell, col: Any=col) -> IEncodable:
        return encode_no_number(cell)

    values: IEnumerable_1[tuple[str, IEncodable]] = to_enumerable([("cells", seq(map(mapping, col)))])
    class ObjectExpr351(IEncodable):
        def Encode(self, helpers: IEncoderHelpers_1[Any], col: Any=col) -> Any:
            def mapping_1(tupled_arg: tuple[str, IEncodable]) -> tuple[str, __A_]:
                return (tupled_arg[0], tupled_arg[1].Encode(helpers))

            arg: IEnumerable_1[tuple[str, __A_]] = map(mapping_1, values)
            return helpers.encode_object(arg)

    return ObjectExpr351()


def _arrow353(builder: IGetters) -> tuple[int | None, IEnumerable_1[FsCell]]:
    n: int | None
    object_arg: IOptionalGetter = builder.Optional
    n = object_arg.Field("number", int_1)
    def _arrow352(__unit: None=None) -> IEnumerable_1[FsCell] | None:
        arg_3: Decoder_1[IEnumerable_1[FsCell]] = seq_1(decode_cols(n))
        object_arg_1: IOptionalGetter = builder.Optional
        return object_arg_1.Field("cells", arg_3)

    return (n, default_arg(_arrow352(), empty()))


decode: Decoder_1[tuple[int | None, IEnumerable_1[FsCell]]] = object(_arrow353)

__all__ = ["encode", "encode_no_numbers", "decode"]




from __future__ import annotations
from collections.abc import Callable
from dataclasses import dataclass
from typing import Any
from ...fable_library.option import (default_arg, map, bind)
from ...fable_library.reflection import (TypeInfo, string_type, option_type, record_type)
from ...fable_library.types import (Array, to_string, Record)
from ..comment import Comment
from ..Helper.collections_ import Option_fromValueWithDefault
from ..ontology_annotation import (OntologyAnnotation, OntologyAnnotation_reflection)

def _expr761() -> TypeInfo:
    return record_type("ARCtrl.Process.ProtocolParameter", [], ProtocolParameter, lambda: [("ID", option_type(string_type)), ("ParameterName", option_type(OntologyAnnotation_reflection()))])


@dataclass(eq = False, repr = False, slots = True)
class ProtocolParameter(Record):
    ID: str | None
    ParameterName: OntologyAnnotation | None
    @staticmethod
    def make(id: str | None=None, parameter_name: OntologyAnnotation | None=None) -> ProtocolParameter:
        return ProtocolParameter(id, parameter_name)

    @staticmethod
    def create(Id: str | None=None, ParameterName: OntologyAnnotation | None=None) -> ProtocolParameter:
        return ProtocolParameter.make(Id, ParameterName)

    @staticmethod
    def empty() -> ProtocolParameter:
        return ProtocolParameter.create()

    @staticmethod
    def from_string(term: str, source: str, accession: str, comments: Array[Comment] | None=None) -> ProtocolParameter:
        oa: OntologyAnnotation = OntologyAnnotation.create(term, source, accession, comments)
        parameter_name: OntologyAnnotation | None = Option_fromValueWithDefault(OntologyAnnotation(), oa)
        return ProtocolParameter.make(None, parameter_name)

    @staticmethod
    def to_string_object(pp: ProtocolParameter) -> dict[str, Any]:
        value: dict[str, Any] = {
            "TermAccessionNumber": "",
            "TermName": "",
            "TermSourceREF": ""
        }
        def mapping(oa: OntologyAnnotation) -> dict[str, Any]:
            return OntologyAnnotation.to_string_object(oa)

        return default_arg(map(mapping, pp.ParameterName), value)

    @property
    def NameText(self, __unit: None=None) -> str:
        this: ProtocolParameter = self
        def mapping(oa: OntologyAnnotation) -> str:
            return oa.NameText

        return default_arg(map(mapping, this.ParameterName), "")

    @property
    def TryNameText(self, __unit: None=None) -> str | None:
        this: ProtocolParameter = self
        def binder(oa: OntologyAnnotation) -> str | None:
            return oa.Name

        return bind(binder, this.ParameterName)

    def MapCategory(self, f: Callable[[OntologyAnnotation], OntologyAnnotation]) -> ProtocolParameter:
        this: ProtocolParameter = self
        return ProtocolParameter(this.ID, map(f, this.ParameterName))

    def SetCategory(self, c: OntologyAnnotation) -> ProtocolParameter:
        this: ProtocolParameter = self
        return ProtocolParameter(this.ID, c)

    @staticmethod
    def try_get_name_text(pp: ProtocolParameter) -> str | None:
        return pp.TryNameText

    @staticmethod
    def get_name_text(pp: ProtocolParameter) -> str:
        return default_arg(ProtocolParameter.try_get_name_text(pp), "")

    @staticmethod
    def name_equals_string(name: str, pp: ProtocolParameter) -> bool:
        return pp.NameText == name

    def Print(self, __unit: None=None) -> str:
        this: ProtocolParameter = self
        return to_string(this)

    def PrintCompact(self, __unit: None=None) -> str:
        this: ProtocolParameter = self
        return "OA " + this.NameText


ProtocolParameter_reflection = _expr761

__all__ = ["ProtocolParameter_reflection"]




from __future__ import annotations
from abc import abstractmethod
from typing import (Protocol, Generic, Any, TypeVar)
from ..fable_library.list import FSharpList
from ..fable_library.reflection import (TypeInfo, string_type, tuple_type, list_type, union_type, char_type, float64_type, bool_type, class_type, array_type, uint32_type)
from ..fable_library.result import FSharpResult_2
from ..fable_library.types import (Array, float32, uint32, Union)
from ..fable_library.util import IEnumerable_1

_JSONVALUE = TypeVar("_JSONVALUE")

_T = TypeVar("_T")

class IDecoderHelpers_1(Protocol, Generic[_JSONVALUE]):
    @abstractmethod
    def any_to_string(self, __arg0: _JSONVALUE) -> str:
        ...

    @abstractmethod
    def as_array(self, __arg0: _JSONVALUE) -> Array[_JSONVALUE]:
        ...

    @abstractmethod
    def as_boolean(self, __arg0: _JSONVALUE) -> bool:
        ...

    @abstractmethod
    def as_float(self, __arg0: _JSONVALUE) -> float:
        ...

    @abstractmethod
    def as_float32(self, __arg0: _JSONVALUE) -> float32:
        ...

    @abstractmethod
    def as_int(self, __arg0: _JSONVALUE) -> int:
        ...

    @abstractmethod
    def as_string(self, __arg0: _JSONVALUE) -> str:
        ...

    @abstractmethod
    def get_properties(self, __arg0: _JSONVALUE) -> IEnumerable_1[str]:
        ...

    @abstractmethod
    def get_property(self, __arg0: str, __arg1: _JSONVALUE) -> _JSONVALUE:
        ...

    @abstractmethod
    def has_property(self, __arg0: str, __arg1: _JSONVALUE) -> bool:
        ...

    @abstractmethod
    def is_array(self, __arg0: _JSONVALUE) -> bool:
        ...

    @abstractmethod
    def is_boolean(self, __arg0: _JSONVALUE) -> bool:
        ...

    @abstractmethod
    def is_integral_value(self, __arg0: _JSONVALUE) -> bool:
        ...

    @abstractmethod
    def is_null_value(self, __arg0: _JSONVALUE) -> bool:
        ...

    @abstractmethod
    def is_number(self, __arg0: _JSONVALUE) -> bool:
        ...

    @abstractmethod
    def is_object(self, __arg0: _JSONVALUE) -> bool:
        ...

    @abstractmethod
    def is_string(self, __arg0: _JSONVALUE) -> bool:
        ...


class IEncoderHelpers_1(Protocol, Generic[_JSONVALUE]):
    @abstractmethod
    def encode_array(self, __arg0: Array[_JSONVALUE]) -> _JSONVALUE:
        ...

    @abstractmethod
    def encode_bool(self, __arg0: bool) -> _JSONVALUE:
        ...

    @abstractmethod
    def encode_char(self, __arg0: str) -> _JSONVALUE:
        ...

    @abstractmethod
    def encode_decimal_number(self, __arg0: float) -> _JSONVALUE:
        ...

    @abstractmethod
    def encode_list(self, __arg0: FSharpList[_JSONVALUE]) -> _JSONVALUE:
        ...

    @abstractmethod
    def encode_null(self) -> _JSONVALUE:
        ...

    @abstractmethod
    def encode_object(self, __arg0: IEnumerable_1[tuple[str, _JSONVALUE]]) -> _JSONVALUE:
        ...

    @abstractmethod
    def encode_resize_array(self, __arg0: Array[_JSONVALUE]) -> _JSONVALUE:
        ...

    @abstractmethod
    def encode_seq(self, __arg0: IEnumerable_1[_JSONVALUE]) -> _JSONVALUE:
        ...

    @abstractmethod
    def encode_signed_integral_number(self, __arg0: int) -> _JSONVALUE:
        ...

    @abstractmethod
    def encode_string(self, __arg0: str) -> _JSONVALUE:
        ...

    @abstractmethod
    def encode_unsigned_integral_number(self, __arg0: uint32) -> _JSONVALUE:
        ...


def _expr72(gen0: TypeInfo) -> TypeInfo:
    return union_type("Thoth.Json.Core.ErrorReason`1", [gen0], ErrorReason_1, lambda: [[("Item1", string_type), ("Item2", gen0)], [("Item1", string_type), ("Item2", gen0), ("Item3", string_type)], [("Item1", string_type), ("Item2", gen0)], [("Item1", string_type), ("Item2", gen0)], [("Item1", string_type), ("Item2", gen0), ("Item3", string_type)], [("Item1", string_type), ("Item2", gen0)], [("Item", string_type)], [("Item", list_type(tuple_type(string_type, ErrorReason_1_reflection(gen0))))]])


class ErrorReason_1(Union, Generic[_JSONVALUE]):
    def __init__(self, tag: int, *fields: Any) -> None:
        super().__init__()
        self.tag: int = tag or 0
        self.fields: Array[Any] = list(fields)

    @staticmethod
    def cases() -> list[str]:
        return ["BadPrimitive", "BadPrimitiveExtra", "BadType", "BadField", "BadPath", "TooSmallArray", "FailMessage", "BadOneOf"]


ErrorReason_1_reflection = _expr72

class Decoder_1(Protocol, Generic[_T]):
    @abstractmethod
    def Decode(self, helpers: IDecoderHelpers_1[_JSONVALUE], value: _JSONVALUE) -> FSharpResult_2[_T, tuple[str, ErrorReason_1[_JSONVALUE]]]:
        ...


def _expr73() -> TypeInfo:
    return union_type("Thoth.Json.Core.Json", [], Json, lambda: [[("Item", string_type)], [("Item", char_type)], [("Item", float64_type)], [], [("Item", bool_type)], [("Item", class_type("System.Collections.Generic.IEnumerable`1", [tuple_type(string_type, Json_reflection())]))], [("Item", array_type(Json_reflection()))], [("Item", uint32_type)], []])


class Json(Union):
    def __init__(self, tag: int, *fields: Any) -> None:
        super().__init__()
        self.tag: int = tag or 0
        self.fields: Array[Any] = list(fields)

    @staticmethod
    def cases() -> list[str]:
        return ["String", "Char", "DecimalNumber", "Null", "Boolean", "Object", "Array", "IntegralNumber", "Unit"]


Json_reflection = _expr73

class IEncodable(Protocol):
    @abstractmethod
    def Encode(self, helpers: IEncoderHelpers_1[_JSONVALUE]) -> _JSONVALUE:
        ...


__all__ = ["ErrorReason_1_reflection", "Json_reflection"]




from __future__ import annotations
from collections.abc import Callable
from typing import (Any, TypeVar)
from ...arctrl_core.Helper.collections_ import (Dictionary_items, Dictionary_tryFind)
from ...arctrl_core.ontology_annotation import OntologyAnnotation
from ...fable_library.array_ import map as map_1
from ...fable_library.map_util import add_to_dict
from ...fable_library.seq import (to_array, map, sort_by)
from ...fable_library.types import Array
from ...fable_library.util import compare_primitives
from ...thoth_json_core.decode import (array as array_2, object, int_1, IGetters)
from ...thoth_json_core.types import (IEncodable, IEncoderHelpers_1, Decoder_1)
from ..ontology_annotation import (OntologyAnnotation_compressedEncoder, OntologyAnnotation_compressedDecoder)

__A_ = TypeVar("__A_")

def array_from_map(otm: Any) -> Array[OntologyAnnotation]:
    def mapping(kv_1: Any, otm: Any=otm) -> OntologyAnnotation:
        return kv_1[0]

    def projection(kv: Any, otm: Any=otm) -> int:
        return kv[1]

    class ObjectExpr2737:
        @property
        def Compare(self) -> Callable[[int, int], int]:
            return compare_primitives

    return to_array(map(mapping, sort_by(projection, Dictionary_items(otm), ObjectExpr2737())))


def encoder(string_table: Any, ot: Array[OntologyAnnotation]) -> IEncodable:
    def mapping(oa: OntologyAnnotation, string_table: Any=string_table, ot: Any=ot) -> IEncodable:
        return OntologyAnnotation_compressedEncoder(string_table, oa)

    values: Array[IEncodable] = map_1(mapping, ot, None)
    class ObjectExpr2738(IEncodable):
        def Encode(self, helpers: IEncoderHelpers_1[Any], string_table: Any=string_table, ot: Any=ot) -> Any:
            def mapping_1(v: IEncodable) -> __A_:
                return v.Encode(helpers)

            arg: Array[__A_] = map_1(mapping_1, values, None)
            return helpers.encode_array(arg)

    return ObjectExpr2738()


def decoder(string_table: Array[str]) -> Decoder_1[Array[OntologyAnnotation]]:
    return array_2(OntologyAnnotation_compressedDecoder(string_table))


def encode_oa(otm: Any, oa: OntologyAnnotation) -> IEncodable:
    match_value: int | None = Dictionary_tryFind(oa, otm)
    if match_value is None:
        i_1: int = len(otm) or 0
        add_to_dict(otm, oa, i_1)
        class ObjectExpr2739(IEncodable):
            def Encode(self, helpers_1: IEncoderHelpers_1[Any], otm: Any=otm, oa: Any=oa) -> Any:
                return helpers_1.encode_signed_integral_number(i_1)

        return ObjectExpr2739()

    else: 
        i: int = match_value or 0
        class ObjectExpr2740(IEncodable):
            def Encode(self, helpers: IEncoderHelpers_1[Any], otm: Any=otm, oa: Any=oa) -> Any:
                return helpers.encode_signed_integral_number(i)

        return ObjectExpr2740()



def decode_oa(ot: Array[OntologyAnnotation]) -> Decoder_1[OntologyAnnotation]:
    def _arrow2741(get: IGetters, ot: Any=ot) -> OntologyAnnotation:
        i: int = get.Required.Raw(int_1) or 0
        return ot[i]

    return object(_arrow2741)


__all__ = ["array_from_map", "encoder", "decoder", "encode_oa", "decode_oa"]




from __future__ import annotations
from collections.abc import Callable
from datetime import datetime
from typing import (Any, TypeVar)
from ...arctrl_core.Helper.collections_ import ResizeArray_map
from ...fable_library.option import (value, default_arg)
from ...fable_library.reflection import (TypeInfo, class_type)
from ...fable_library.types import Array
from ..helper import clean
from ..ldcontext import LDContext
from ..ldobject import (LDNode, LDGraph)
from .file import LDFile
from .lab_protocol import LDLabProtocol
from .person import LDPerson
from .property_value import LDPropertyValue
from .sample import LDSample

__C = TypeVar("__C")

__B = TypeVar("__B")

def _expr1530() -> TypeInfo:
    return class_type("ARCtrl.ROCrate.LDLabProcess", None, LDLabProcess)


class LDLabProcess:
    @staticmethod
    def schema_type() -> str:
        return "https://bioschemas.org/LabProcess"

    @staticmethod
    def name() -> str:
        return "http://schema.org/name"

    @staticmethod
    def agent() -> str:
        return "http://schema.org/agent"

    @staticmethod
    def object_() -> str:
        return "http://schema.org/object"

    @staticmethod
    def result() -> str:
        return "http://schema.org/result"

    @staticmethod
    def executes_lab_protocol() -> str:
        return "https://bioschemas.org/properties/executesLabProtocol"

    @staticmethod
    def executes_lab_protocol_deprecated() -> str:
        return "https://bioschemas.org/executesLabProtocol"

    @staticmethod
    def parameter_value() -> str:
        return "https://bioschemas.org/properties/parameterValue"

    @staticmethod
    def parameter_value_deprecated() -> str:
        return "https://bioschemas.org/parameterValue"

    @staticmethod
    def end_time() -> str:
        return "http://schema.org/endTime"

    @staticmethod
    def disambiguating_description() -> str:
        return "http://schema.org/disambiguatingDescription"

    @staticmethod
    def try_get_name_as_string(lp: LDNode, context: LDContext | None=None) -> str | None:
        match_value: Any | None = lp.TryGetPropertyAsSingleton(LDLabProcess.name(), context)
        (pattern_matching_result, n) = (None, None)
        if match_value is not None:
            if str(type(value(match_value))) == "<class \'str\'>":
                pattern_matching_result = 0
                n = value(match_value)

            else: 
                pattern_matching_result = 1


        else: 
            pattern_matching_result = 1

        if pattern_matching_result == 0:
            return n

        elif pattern_matching_result == 1:
            return None


    @staticmethod
    def get_name_as_string(lp: LDNode, context: LDContext | None=None) -> str:
        match_value: Any | None = lp.TryGetPropertyAsSingleton(LDLabProcess.name(), context)
        if match_value is not None:
            if str(type(value(match_value))) == "<class \'str\'>":
                n: str = value(match_value)
                return n

            else: 
                raise Exception(("property `name` of object with @id `" + lp.Id) + "` was not a string")


        else: 
            raise Exception(("Could not access property `name` of object with @id `" + lp.Id) + "`")


    @staticmethod
    def set_name_as_string(lp: LDNode, name: str, context: LDContext | None=None) -> Any:
        return lp.SetProperty(LDLabProcess.name(), name, context)

    @staticmethod
    def try_get_agent(lp: LDNode, graph: LDGraph | None=None, context: LDContext | None=None) -> LDNode | None:
        match_value: LDNode | None = lp.TryGetPropertyAsSingleNode(LDLabProcess.agent(), graph, context)
        (pattern_matching_result, a_1) = (None, None)
        if match_value is not None:
            def _arrow1525(__unit: None=None) -> bool:
                ld_object: LDNode = match_value
                return LDPerson.validate(ld_object, context)

            if _arrow1525():
                pattern_matching_result = 0
                a_1 = match_value

            else: 
                pattern_matching_result = 1


        else: 
            pattern_matching_result = 1

        if pattern_matching_result == 0:
            return a_1

        elif pattern_matching_result == 1:
            return None


    @staticmethod
    def get_agent(lp: LDNode, graph: LDGraph | None=None, context: LDContext | None=None) -> LDNode:
        match_value: LDNode | None = lp.TryGetPropertyAsSingleNode(LDLabProcess.agent(), graph, context)
        if match_value is not None:
            def _arrow1526(__unit: None=None) -> bool:
                ld_object: LDNode = match_value
                return LDPerson.validate(ld_object, context)

            if _arrow1526():
                a_1: LDNode = match_value
                return a_1

            else: 
                raise Exception(("Property of `agent` of object with @id `" + lp.Id) + "` was not a valid Person")


        else: 
            raise Exception(("Could not access property `agent` of object with @id `" + lp.Id) + "`")


    @staticmethod
    def set_agent(lp: LDNode, agent: LDNode, context: LDContext | None=None) -> Any:
        return lp.SetProperty(LDLabProcess.agent(), agent, context)

    @staticmethod
    def get_objects(lp: LDNode, graph: LDGraph | None=None, context: LDContext | None=None) -> Array[LDNode]:
        return lp.GetPropertyNodes(LDLabProcess.object_(), None, graph, context)

    @staticmethod
    def get_objects_as_sample(lp: LDNode, graph: LDGraph | None=None, context: LDContext | None=None) -> Array[LDNode]:
        def filter(ld_object: LDNode, context_1: LDContext | None=None) -> bool:
            return LDSample.validate(ld_object, context_1)

        return lp.GetPropertyNodes(LDLabProcess.object_(), filter, graph, context)

    @staticmethod
    def get_objects_as_data(lp: LDNode, graph: LDGraph | None=None, context: LDContext | None=None) -> Array[LDNode]:
        def filter(ld_object: LDNode, context_1: LDContext | None=None) -> bool:
            return LDFile.validate(ld_object, context_1)

        return lp.GetPropertyNodes(LDLabProcess.object_(), filter, graph, context)

    @staticmethod
    def set_objects(lp: LDNode, objects: Array[LDNode], context: LDContext | None=None) -> Any:
        return lp.SetProperty(LDLabProcess.object_(), objects, context)

    @staticmethod
    def get_results(lp: LDNode, graph: LDGraph | None=None, context: LDContext | None=None) -> Array[LDNode]:
        return lp.GetPropertyNodes(LDLabProcess.result(), None, graph, context)

    @staticmethod
    def get_results_as_sample(lp: LDNode, graph: LDGraph | None=None, context: LDContext | None=None) -> Array[LDNode]:
        def filter(ld_object: LDNode, context_1: LDContext | None=None) -> bool:
            return LDSample.validate(ld_object, context_1)

        return lp.GetPropertyNodes(LDLabProcess.result(), filter, graph, context)

    @staticmethod
    def get_results_as_data(lp: LDNode, graph: LDGraph | None=None, context: LDContext | None=None) -> Array[LDNode]:
        def filter(ld_object: LDNode, context_1: LDContext | None=None) -> bool:
            return LDFile.validate(ld_object, context_1)

        return lp.GetPropertyNodes(LDLabProcess.result(), filter, graph, context)

    @staticmethod
    def set_results(lp: LDNode, results: Array[LDNode], context: LDContext | None=None) -> Any:
        return lp.SetProperty(LDLabProcess.result(), results, context)

    @staticmethod
    def try_get_executes_lab_protocol(lp: LDNode, graph: LDGraph | None=None, context: LDContext | None=None) -> LDNode | None:
        def filter(ld_object: LDNode, context_1: LDContext | None=None) -> bool:
            return LDLabProtocol.validate(ld_object, context_1)

        match_value: LDNode | None = lp.TryGetPropertyAsSingleNode(LDLabProcess.executes_lab_protocol(), graph, context)
        (pattern_matching_result, l_1) = (None, None)
        if match_value is not None:
            if filter(match_value, context):
                pattern_matching_result = 0
                l_1 = match_value

            else: 
                pattern_matching_result = 1


        else: 
            pattern_matching_result = 1

        if pattern_matching_result == 0:
            return l_1

        elif pattern_matching_result == 1:
            match_value_1: LDNode | None = lp.TryGetPropertyAsSingleNode(LDLabProcess.executes_lab_protocol_deprecated(), graph, context)
            (pattern_matching_result_1, l_3) = (None, None)
            if match_value_1 is not None:
                if filter(match_value_1, context):
                    pattern_matching_result_1 = 0
                    l_3 = match_value_1

                else: 
                    pattern_matching_result_1 = 1


            else: 
                pattern_matching_result_1 = 1

            if pattern_matching_result_1 == 0:
                return l_3

            elif pattern_matching_result_1 == 1:
                return None



    @staticmethod
    def set_executes_lab_protocol(lp: LDNode, executes_lab_protocol: LDNode, context: LDContext | None=None) -> Any:
        return lp.SetProperty(LDLabProcess.executes_lab_protocol(), executes_lab_protocol, context)

    @staticmethod
    def get_parameter_values(lp: LDNode, graph: LDGraph | None=None, context: LDContext | None=None) -> Array[LDNode]:
        def filter(ld_object: LDNode, context_1: LDContext | None=None) -> bool:
            return LDPropertyValue.validate(ld_object, context_1)

        l: Array[LDNode] = lp.GetPropertyNodes(LDLabProcess.parameter_value(), filter, graph, context)
        return lp.GetPropertyNodes(LDLabProcess.parameter_value_deprecated(), filter, graph, context) if (len(l) == 0) else l

    @staticmethod
    def set_parameter_values(lp: LDNode, parameter_values: Array[LDNode], context: LDContext | None=None) -> Any:
        return lp.SetProperty(LDLabProcess.parameter_value(), parameter_values, context)

    @staticmethod
    def try_get_end_time(lp: LDNode, context: LDContext | None=None) -> Any | None:
        match_value: Any | None = lp.TryGetPropertyAsSingleton(LDLabProcess.end_time(), context)
        (pattern_matching_result, et) = (None, None)
        if match_value is not None:
            if isinstance(value(match_value), datetime):
                pattern_matching_result = 0
                et = value(match_value)

            else: 
                pattern_matching_result = 1


        else: 
            pattern_matching_result = 1

        if pattern_matching_result == 0:
            return et

        elif pattern_matching_result == 1:
            return None


    @staticmethod
    def set_end_time(lp: LDNode, end_time: Any, context: LDContext | None=None) -> Any:
        return lp.SetProperty(LDLabProcess.end_time(), end_time, context)

    @staticmethod
    def get_disambiguating_descriptions_as_string(lp: LDNode, context: LDContext | None=None) -> Array[str]:
        def f(o_1: Any=None) -> Any:
            return o_1

        def filter(o: Any=None, context_1: LDContext | None=None) -> bool:
            return str(type(o)) == "<class \'str\'>"

        return ResizeArray_map(f, lp.GetPropertyValues(LDLabProcess.disambiguating_description(), filter, context))

    @staticmethod
    def set_disambiguating_descriptions_as_string(lp: LDNode, disambiguating_descriptions: Array[str], context: LDContext | None=None) -> Any:
        return lp.SetProperty(LDLabProcess.disambiguating_description(), disambiguating_descriptions, context)

    @staticmethod
    def validate(lp: LDNode, context: LDContext | None=None) -> bool:
        return lp.HasProperty(LDLabProcess.name(), context) if lp.HasType(LDLabProcess.schema_type(), context) else False

    @staticmethod
    def gen_id(name: Any, assay_name: Any | None=None, study_name: Any | None=None) -> str:
        def _arrow1527(__unit: None=None) -> str:
            study_1: __C = value(study_name)
            return ((("#Process_S_" + str(study_1)) + "_") + str(name)) + ""

        def _arrow1528(__unit: None=None) -> str:
            assay_1: __B = value(assay_name)
            return ((("#Process_A_" + str(assay_1)) + "_") + str(name)) + ""

        def _arrow1529(__unit: None=None) -> str:
            assay: __B = value(assay_name)
            study: __C = value(study_name)
            return ((((("#Process_S_" + str(study)) + "_A_") + str(assay)) + "_") + str(name)) + ""

        return clean((_arrow1527() if (study_name is not None) else (("#Process_" + str(name)) + "")) if (assay_name is None) else (_arrow1528() if (study_name is None) else _arrow1529()))

    @staticmethod
    def create(name: str, objects: Array[LDNode] | None=None, results: Array[LDNode] | None=None, id: str | None=None, agent: LDNode | None=None, executes_lab_protocol: LDNode | None=None, parameter_values: Array[LDNode] | None=None, end_time: Any | None=None, disambiguating_descriptions: Array[str] | None=None, context: LDContext | None=None) -> LDNode:
        id_1: str = LDLabProcess.gen_id(name) if (id is None) else id
        objects_1: Array[LDNode] = default_arg(objects, [])
        results_1: Array[LDNode] = default_arg(results, [])
        lp: LDNode = LDNode(id_1, [LDLabProcess.schema_type()], None, context)
        lp.SetProperty(LDLabProcess.name(), name, context)
        lp.SetOptionalProperty(LDLabProcess.agent(), agent, context)
        lp.SetProperty(LDLabProcess.object_(), objects_1, context)
        lp.SetProperty(LDLabProcess.result(), results_1, context)
        lp.SetOptionalProperty(LDLabProcess.executes_lab_protocol(), executes_lab_protocol, context)
        lp.SetOptionalProperty(LDLabProcess.parameter_value(), parameter_values, context)
        lp.SetOptionalProperty(LDLabProcess.end_time(), end_time, context)
        lp.SetOptionalProperty(LDLabProcess.disambiguating_description(), disambiguating_descriptions, context)
        return lp


LDLabProcess_reflection = _expr1530

__all__ = ["LDLabProcess_reflection"]




from __future__ import annotations
from typing import Any
from ..fable_library.option import (default_arg, value as value_4)
from ..fable_library.reflection import (TypeInfo, class_type)
from ..fable_library.seq import (to_list, delay, append, singleton, empty)
from ..fable_library.string_ import (join, to_text, printf)
from ..fable_library.system_text import (StringBuilder__ctor, StringBuilder__Append_Z721C83C5)
from ..fable_library.types import (Array, to_string)
from ..fable_library.util import (ignore, IEnumerable_1, string_hash, safe_hash)
from .comment import Comment
from .Helper.collections_ import ResizeArray_map
from .Helper.hash_codes import (merge_hashes, hash_1 as hash_1_1)
from .Helper.regex import (try_parse_term_annotation, ActivePatterns__007CRegex_007C__007C, Pattern_TermAnnotationShortPattern)
from .Helper.url import create_oauri

def _expr722() -> TypeInfo:
    return class_type("ARCtrl.OntologyAnnotation", None, OntologyAnnotation)


class OntologyAnnotation:
    def __init__(self, name: str | None=None, tsr: str | None=None, tan: str | None=None, comments: Array[Comment] | None=None) -> None:
        self._name: str | None = None if (name is None) else (None if (name == "") else name)
        self._termSourceREF: str | None = None if (tsr is None) else (None if (tsr == "") else tsr)
        self._termAccessionNumber: str | None = None if (tan is None) else (None if (tan == "") else (None if (tan == ":") else tan))
        self._comments: Array[Comment] = default_arg(comments, [])
        self._tanInfo: dict[str, Any] | None = OntologyAnnotation.compute_tan_info(self._termAccessionNumber, self._termSourceREF)

    @property
    def Name(self, __unit: None=None) -> str | None:
        this: OntologyAnnotation = self
        return this._name

    @Name.setter
    def Name(self, name: str | None=None) -> None:
        this: OntologyAnnotation = self
        this._name = name

    @property
    def TermSourceREF(self, __unit: None=None) -> str | None:
        this: OntologyAnnotation = self
        return this._termSourceREF

    @TermSourceREF.setter
    def TermSourceREF(self, tsr: str | None=None) -> None:
        this: OntologyAnnotation = self
        this._termSourceREF = tsr
        this.recompute_tan_info()

    @property
    def TermAccessionNumber(self, __unit: None=None) -> str | None:
        this: OntologyAnnotation = self
        return this._termAccessionNumber

    @TermAccessionNumber.setter
    def TermAccessionNumber(self, tan: str | None=None) -> None:
        this: OntologyAnnotation = self
        this._termAccessionNumber = tan
        this.recompute_tan_info()

    @property
    def Comments(self, __unit: None=None) -> Array[Comment]:
        this: OntologyAnnotation = self
        return this._comments

    @Comments.setter
    def Comments(self, comments: Array[Comment]) -> None:
        this: OntologyAnnotation = self
        this._comments = comments

    @staticmethod
    def make(name: str | None, tsr: str | None, tan: str | None, comments: Array[Comment]) -> OntologyAnnotation:
        return OntologyAnnotation(name, tsr, tan, comments)

    @staticmethod
    def create(name: str | None=None, tsr: str | None=None, tan: str | None=None, comments: Array[Comment] | None=None) -> OntologyAnnotation:
        comments_1: Array[Comment] = default_arg(comments, [])
        return OntologyAnnotation.make(name, tsr, tan, comments_1)

    @property
    def TANInfo(self, __unit: None=None) -> dict[str, Any] | None:
        this: OntologyAnnotation = self
        return this._tanInfo

    @property
    def NameText(self, __unit: None=None) -> str:
        this: OntologyAnnotation = self
        return default_arg(this.Name, "")

    @staticmethod
    def create_uri_annotation(term_source_ref: str, local_tan: str) -> str:
        return create_oauri(term_source_ref, local_tan)

    @staticmethod
    def from_term_annotation(tan: str, name: str | None=None) -> OntologyAnnotation:
        match_value: dict[str, Any] | None = try_parse_term_annotation(tan)
        if match_value is None:
            return OntologyAnnotation.create(name, None, tan)

        else: 
            r: dict[str, Any] = match_value
            accession: str = (r["IDSpace"] + ":") + r["LocalID"]
            return OntologyAnnotation.create(name, r["IDSpace"], accession)


    @property
    def TermAccessionShort(self, __unit: None=None) -> str:
        this: OntologyAnnotation = self
        match_value: dict[str, Any] | None = this.TANInfo
        if match_value is not None:
            id: dict[str, Any] = match_value
            return ((("" + id["IDSpace"]) + ":") + id["LocalID"]) + ""

        else: 
            return ""


    @property
    def TermAccessionOntobeeUrl(self, __unit: None=None) -> str:
        this: OntologyAnnotation = self
        match_value: dict[str, Any] | None = this.TANInfo
        if match_value is not None:
            id: dict[str, Any] = match_value
            return OntologyAnnotation.create_uri_annotation(id["IDSpace"], id["LocalID"])

        else: 
            return ""


    @property
    def TermAccessionAndOntobeeUrlIfShort(self, __unit: None=None) -> str:
        this: OntologyAnnotation = self
        match_value: str | None = this.TermAccessionNumber
        if match_value is not None:
            tan: str = match_value
            return this.TermAccessionOntobeeUrl if (ActivePatterns__007CRegex_007C__007C(Pattern_TermAnnotationShortPattern, tan) is not None) else tan

        else: 
            return ""


    @staticmethod
    def to_string_object(oa: OntologyAnnotation, as_ontobee_purl_url_if_short: bool | None=None) -> dict[str, Any]:
        as_ontobee_purl_url_if_short_1: bool = default_arg(as_ontobee_purl_url_if_short, False)
        TermName: str = default_arg(oa.Name, "")
        TermSourceREF: str = default_arg(oa.TermSourceREF, "")
        def _arrow709(__unit: None=None) -> str:
            url: str = oa.TermAccessionAndOntobeeUrlIfShort
            return default_arg(oa.TermAccessionNumber, "") if (url == "") else url

        return {
            "TermAccessionNumber": _arrow709() if as_ontobee_purl_url_if_short_1 else default_arg(oa.TermAccessionNumber, ""),
            "TermName": TermName,
            "TermSourceREF": TermSourceREF
        }

    def __str__(self, __unit: None=None) -> str:
        this: OntologyAnnotation = self
        sb: Any = StringBuilder__ctor()
        ignore(StringBuilder__Append_Z721C83C5(sb, "{"))
        def _arrow717(__unit: None=None) -> IEnumerable_1[str]:
            def _arrow710(__unit: None=None) -> str:
                arg: str = value_4(this.Name)
                return to_text(printf("Name = %A"))(arg)

            def _arrow716(__unit: None=None) -> IEnumerable_1[str]:
                def _arrow711(__unit: None=None) -> str:
                    arg_1: str = value_4(this.TermSourceREF)
                    return to_text(printf("TSR = %A"))(arg_1)

                def _arrow715(__unit: None=None) -> IEnumerable_1[str]:
                    def _arrow712(__unit: None=None) -> str:
                        arg_2: str = value_4(this.TermAccessionNumber)
                        return to_text(printf("TAN = %A"))(arg_2)

                    def _arrow714(__unit: None=None) -> IEnumerable_1[str]:
                        def _arrow713(__unit: None=None) -> str:
                            arg_3: Array[Comment] = this.Comments
                            return to_text(printf("Comments = %A"))(arg_3)

                        return singleton(_arrow713()) if (len(this.Comments) != 0) else empty()

                    return append(singleton(_arrow712()) if (this.TermAccessionNumber is not None) else empty(), delay(_arrow714))

                return append(singleton(_arrow711()) if (this.TermSourceREF is not None) else empty(), delay(_arrow715))

            return append(singleton(_arrow710()) if (this.Name is not None) else empty(), delay(_arrow716))

        ignore(StringBuilder__Append_Z721C83C5(sb, join("; ", to_list(delay(_arrow717)))))
        ignore(StringBuilder__Append_Z721C83C5(sb, "}"))
        return to_string(sb)

    def is_empty(self, __unit: None=None) -> bool:
        this: OntologyAnnotation = self
        return (len(this.Comments) == 0) if ((this.TermAccessionNumber is None) if ((this.TermSourceREF is None) if (this.Name is None) else False) else False) else False

    def __hash__(self, __unit: None=None) -> int:
        this: OntologyAnnotation = self
        def _arrow718(__unit: None=None) -> int:
            match_value: str | None = this.Name
            return 0 if (match_value is None) else string_hash(match_value)

        def _arrow720(__unit: None=None) -> int:
            match_value: str | None = this.TermSourceREF
            match_value_1: dict[str, Any] | None = this.TANInfo
            if match_value is not None:
                if match_value_1 is None:
                    tsr_1: str = match_value
                    match_value_2: str | None = this.TermAccessionNumber
                    if match_value_2 is None:
                        return string_hash(tsr_1.lower())

                    else: 
                        tan: str = match_value_2
                        return merge_hashes(string_hash(tsr_1.lower()), string_hash(tan))


                else: 
                    taninfo_1: dict[str, Any] = match_value_1
                    tsr: str = match_value
                    return merge_hashes(string_hash(tsr.lower()), string_hash(taninfo_1["LocalID"]))


            else: 
                def _arrow719(__unit: None=None) -> int:
                    taninfo: dict[str, Any] = match_value_1
                    return merge_hashes(string_hash(taninfo["IDSpace"].lower()), string_hash(taninfo["LocalID"]))

                return 0 if (match_value_1 is None) else _arrow719()


        return merge_hashes(_arrow718(), _arrow720())

    def __eq__(self, obj: Any=None) -> bool:
        this: OntologyAnnotation = self
        return hash_1_1(this) == hash_1_1(obj)

    def Copy(self, __unit: None=None) -> OntologyAnnotation:
        this: OntologyAnnotation = self
        def f(c: Comment) -> Comment:
            return c.Copy()

        next_comments: Array[Comment] = ResizeArray_map(f, this.Comments)
        name: str | None = this.Name
        tsr: str | None = this.TermSourceREF
        tan: str | None = this.TermAccessionNumber
        return OntologyAnnotation.make(name, tsr, tan, next_comments)

    @staticmethod
    def compute_tan_info(tan: str | None=None, tsr: str | None=None) -> dict[str, Any] | None:
        if tan is None:
            return None

        else: 
            tan_1: str = tan
            match_value: dict[str, Any] | None = try_parse_term_annotation(tan_1)
            def _arrow721(__unit: None=None) -> dict[str, Any] | None:
                tsr_1: str = tsr
                return {
                    "IDSpace": tsr_1,
                    "LocalID": tan_1
                }

            return (None if (tsr is None) else (None if (tsr == "") else _arrow721())) if (match_value is None) else match_value


    def Print(self, __unit: None=None) -> str:
        this: OntologyAnnotation = self
        return to_string(this)

    def PrintCompact(self, __unit: None=None) -> str:
        this: OntologyAnnotation = self
        return "OA " + this.NameText

    def __cmp__(self, obj: Any=None) -> int:
        this: OntologyAnnotation = self
        if isinstance(obj, OntologyAnnotation):
            hash_1: int = safe_hash(this) or 0
            other_hash: int = safe_hash(obj) or 0
            return 0 if (hash_1 == other_hash) else (-1 if (hash_1 < other_hash) else 1)

        else: 
            return 1


    def recompute_tan_info(self, __unit: None=None) -> None:
        this: OntologyAnnotation = self
        new_tan_info: dict[str, Any] | None = OntologyAnnotation.compute_tan_info(this._termAccessionNumber, this._termSourceREF)
        this._tanInfo = new_tan_info


OntologyAnnotation_reflection = _expr722

def OntologyAnnotation__ctor_Z54349580(name: str | None=None, tsr: str | None=None, tan: str | None=None, comments: Array[Comment] | None=None) -> OntologyAnnotation:
    return OntologyAnnotation(name, tsr, tan, comments)


__all__ = ["OntologyAnnotation_reflection"]




from __future__ import annotations
from collections.abc import Callable
from typing import (Any, TypeVar)
from ....arctrl_core.Helper.collections_ import Dictionary_init
from ....arctrl_core.ontology_annotation import OntologyAnnotation
from ....arctrl_core.Table.composite_cell import CompositeCell
from ....arctrl_json.string_table import (encoder as encoder_3, array_from_map as array_from_map_2, decoder as decoder_1)
from ....arctrl_json.Table.cell_table import (encoder as encoder_1, array_from_map, decoder as decoder_3)
from ....arctrl_json.Table.oatable import (encoder as encoder_2, array_from_map as array_from_map_1, decoder as decoder_2)
from ....fable_library.seq import map
from ....fable_library.types import Array
from ....fable_library.util import (ignore, to_enumerable, IEnumerable_1)
from ....thoth_json_core.decode import (object as object_1, IRequiredGetter, IGetters)
from ....thoth_json_core.types import (IEncodable, IEncoderHelpers_1, Decoder_1)
from ....thoth_json_python.encode import to_string

__A_ = TypeVar("__A_")

_A = TypeVar("_A")

__A = TypeVar("__A")

__B = TypeVar("__B")

def encode(encoder: Callable[[Any, Any, Any, _A], IEncodable], obj: Any) -> IEncodable:
    string_table: Any = Dictionary_init()
    oa_table: Any = Dictionary_init()
    cell_table: Any = Dictionary_init()
    object: IEncodable = encoder(string_table, oa_table, cell_table, obj)
    ignore(to_string(0, object))
    encoded_cell_table: IEncodable = encoder_1(string_table, oa_table, array_from_map(cell_table))
    encoded_oatable: IEncodable = encoder_2(string_table, array_from_map_1(oa_table))
    encoded_string_table: IEncodable = encoder_3(array_from_map_2(string_table))
    class ObjectExpr3836(IEncodable):
        def Encode(self, helpers: IEncoderHelpers_1[Any], encoder: Any=encoder, obj: Any=obj) -> Any:
            def mapping(tupled_arg: tuple[str, IEncodable]) -> tuple[str, __A_]:
                return (tupled_arg[0], tupled_arg[1].Encode(helpers))

            arg: IEnumerable_1[tuple[str, __A_]] = map(mapping, to_enumerable([("cellTable", encoded_cell_table), ("oaTable", encoded_oatable), ("stringTable", encoded_string_table), ("object", object)]))
            return helpers.encode_object(arg)

    return ObjectExpr3836()


def decode(decoder: Callable[[Array[str], Array[OntologyAnnotation], Array[CompositeCell]], __A]) -> Decoder_1[Any]:
    def _arrow3838(get: IGetters, decoder: Any=decoder) -> __B:
        string_table: Array[str]
        object_arg: IRequiredGetter = get.Required
        string_table = object_arg.Field("stringTable", decoder_1)
        oa_table: Array[OntologyAnnotation]
        arg_3: Decoder_1[Array[OntologyAnnotation]] = decoder_2(string_table)
        object_arg_1: IRequiredGetter = get.Required
        oa_table = object_arg_1.Field("oaTable", arg_3)
        def _arrow3837(__unit: None=None) -> Array[CompositeCell]:
            arg_5: Decoder_1[Array[CompositeCell]] = decoder_3(string_table, oa_table)
            object_arg_2: IRequiredGetter = get.Required
            return object_arg_2.Field("cellTable", arg_5)

        arg_7: __A = decoder(string_table, oa_table, _arrow3837())
        object_arg_3: IRequiredGetter = get.Required
        return object_arg_3.Field("object", arg_7)

    return object_1(_arrow3838)


__all__ = ["encode", "decode"]




from __future__ import annotations
from collections.abc import Callable
from typing import Any
from ...fable_library.map_util import (get_item_from_dict, add_to_dict, remove_from_dict, add_to_set)
from ...fable_library.option import (some, value as value_6)
from ...fable_library.range import range_big_int
from ...fable_library.reflection import (TypeInfo, class_type)
from ...fable_library.seq import (map, iterate, is_empty, max, collect, delay, empty, singleton, min, min_by)
from ...fable_library.util import (IEnumerable_1, ignore, compare_primitives)
from ..fs_address import FsAddress
from .fs_cell import FsCell

def Dictionary_tryGet(k: Any, dict_1: Any) -> Any | None:
    if k in dict_1:
        return some(get_item_from_dict(dict_1, k))

    else: 
        return None



def _expr215() -> TypeInfo:
    return class_type("FsSpreadsheet.FsCellsCollection", None, FsCellsCollection)


class FsCellsCollection:
    def __init__(self, __unit: None=None) -> None:
        self._columnsUsed: Any = dict([])
        self._deleted: Any = dict([])
        self._rowsCollection: Any = dict([])
        self._maxColumnUsed: int = 0
        self._maxRowUsed: int = 0
        self._rowsUsed: Any = dict([])
        self._count: int = 0

    @property
    def Count(self, __unit: None=None) -> int:
        this: FsCellsCollection = self
        return this._count

    @Count.setter
    def Count(self, count: int) -> None:
        this: FsCellsCollection = self
        this._count = count or 0

    @property
    def MaxRowNumber(self, __unit: None=None) -> int:
        this: FsCellsCollection = self
        return this._maxRowUsed

    @property
    def MaxColumnNumber(self, __unit: None=None) -> int:
        this: FsCellsCollection = self
        return this._maxColumnUsed

    def Copy(self, __unit: None=None) -> FsCellsCollection:
        this: FsCellsCollection = self
        def mapping(c: FsCell) -> FsCell:
            return c.Copy()

        cells: IEnumerable_1[FsCell] = map(mapping, this.GetCells())
        return FsCellsCollection.create_from_cells(cells)

    @staticmethod
    def copy(cells_collection: FsCellsCollection) -> FsCellsCollection:
        return cells_collection.Copy()

    @staticmethod
    def IncrementUsage(dictionary: Any, key: int) -> None:
        match_value: int | None = Dictionary_tryGet(key, dictionary)
        if match_value is None:
            add_to_dict(dictionary, key, 1)

        else: 
            count: int = match_value or 0
            dictionary[key] = count + 1


    @staticmethod
    def DecrementUsage(dictionary: Any, key: int) -> bool:
        match_value: int | None = Dictionary_tryGet(key, dictionary)
        if match_value is None:
            return False

        elif match_value > 1:
            count_1: int = match_value or 0
            dictionary[key] = count_1 - 1
            return False

        else: 
            ignore(remove_from_dict(dictionary, key))
            return True


    @staticmethod
    def create_from_cells(cells: IEnumerable_1[FsCell]) -> FsCellsCollection:
        fcc: FsCellsCollection = FsCellsCollection()
        fcc.AddMany(cells)
        return fcc

    def Clear(self, __unit: None=None) -> FsCellsCollection:
        this: FsCellsCollection = self
        this._count = 0
        this._rowsUsed.clear()
        this._columnsUsed.clear()
        this._rowsCollection.clear()
        this._maxRowUsed = 0
        this._maxColumnUsed = 0
        return this

    def Add(self, cell: FsCell, row: int | None=None, column: int | None=None) -> None:
        this: FsCellsCollection = self
        if row is not None:
            cell.RowNumber = value_6(row) or 0

        if column is not None:
            cell.ColumnNumber = value_6(column) or 0

        row_1: int = cell.RowNumber or 0
        column_1: int = cell.ColumnNumber or 0
        this._count = (this._count + 1) or 0
        FsCellsCollection.IncrementUsage(this._rowsUsed, row_1)
        FsCellsCollection.IncrementUsage(this._columnsUsed, column_1)
        def _arrow199(__unit: None=None) -> Any:
            match_value: Any | None = Dictionary_tryGet(row_1, this._rowsCollection)
            if match_value is None:
                columns_collection_1: Any = dict([])
                add_to_dict(this._rowsCollection, row_1, columns_collection_1)
                return columns_collection_1

            else: 
                return match_value


        add_to_dict(_arrow199(), column_1, cell)
        if row_1 > this._maxRowUsed:
            this._maxRowUsed = row_1 or 0

        if column_1 > this._maxColumnUsed:
            this._maxColumnUsed = column_1 or 0

        match_value_1: Any | None = Dictionary_tryGet(row_1, this._deleted)
        if match_value_1 is None:
            pass

        else: 
            del_hash: Any = match_value_1
            ignore(del_hash.delete(column_1))


    @staticmethod
    def add_cell_with_indeces(row_index: int, col_index: int, cell: FsCell, cells_collection: FsCellsCollection) -> None:
        cells_collection.Add(cell, row_index, col_index)

    @staticmethod
    def add_cell(cell: FsCell, cells_collection: FsCellsCollection) -> FsCellsCollection:
        cells_collection.Add(cell)
        return cells_collection

    def AddMany(self, cells: IEnumerable_1[FsCell]) -> None:
        this: FsCellsCollection = self
        def action(arg: FsCell) -> None:
            value: None = this.Add(arg)
            ignore(None)

        iterate(action, cells)

    @staticmethod
    def add_cells(cells: IEnumerable_1[FsCell], cells_collection: FsCellsCollection) -> FsCellsCollection:
        cells_collection.AddMany(cells)
        return cells_collection

    def ContainsCellAt(self, row_index: int, col_index: int) -> bool:
        this: FsCellsCollection = self
        match_value: Any | None = Dictionary_tryGet(row_index, this._rowsCollection)
        if match_value is None:
            return False

        else: 
            cols_collection: Any = match_value
            return cols_collection.has(col_index)


    @staticmethod
    def contains_cell_at(row_index: int, col_index: int, cells_collection: FsCellsCollection) -> bool:
        return cells_collection.ContainsCellAt(row_index, col_index)

    def RemoveCellAt(self, row: int, column: int) -> None:
        this: FsCellsCollection = self
        this._count = (this._count - 1) or 0
        row_removed: bool = FsCellsCollection.DecrementUsage(this._rowsUsed, row)
        column_removed: bool = FsCellsCollection.DecrementUsage(this._columnsUsed, column)
        if (row == this._maxRowUsed) if row_removed else False:
            class ObjectExpr200:
                @property
                def Compare(self) -> Callable[[int, int], int]:
                    return compare_primitives

            this._maxRowUsed = (max(this._rowsUsed.keys(), ObjectExpr200()) if (not is_empty(this._rowsUsed.keys())) else 0) or 0

        if (column == this._maxColumnUsed) if column_removed else False:
            class ObjectExpr201:
                @property
                def Compare(self) -> Callable[[int, int], int]:
                    return compare_primitives

            this._maxColumnUsed = (max(this._columnsUsed.keys(), ObjectExpr201()) if (not is_empty(this._columnsUsed.keys())) else 0) or 0

        match_value: Any | None = Dictionary_tryGet(row, this._deleted)
        if match_value is None:
            del_hash_3: Any = set([])
            ignore(add_to_set(column, del_hash_3))
            add_to_dict(this._deleted, row, del_hash_3)

        else: 
            def _arrow202(__unit: None=None) -> bool:
                del_hash: Any = match_value
                return del_hash.has(column)

            if _arrow202():
                del_hash_1: Any = match_value

            else: 
                del_hash_2: Any = match_value
                ignore(add_to_set(column, del_hash_2))


        match_value_1: Any | None = Dictionary_tryGet(row, this._rowsCollection)
        if match_value_1 is None:
            pass

        else: 
            columns_collection: Any = match_value_1
            ignore(remove_from_dict(columns_collection, column))
            if len(columns_collection) == 0:
                ignore(remove_from_dict(this._rowsCollection, row))



    @staticmethod
    def remove_cell_at(row_index: int, col_index: int, cells_collection: FsCellsCollection) -> FsCellsCollection:
        cells_collection.RemoveCellAt(row_index, col_index)
        return cells_collection

    def TryRemoveValueAt(self, row_index: int, col_index: int) -> None:
        this: FsCellsCollection = self
        match_value: Any | None = Dictionary_tryGet(row_index, this._rowsCollection)
        if match_value is None:
            pass

        else: 
            cols_collection: Any = match_value
            try: 
                get_item_from_dict(cols_collection, col_index).Value = ""

            except Exception as match_value_1:
                pass



    @staticmethod
    def try_remove_value_at(row_index: int, col_index: int, cells_collection: FsCellsCollection) -> FsCellsCollection:
        cells_collection.TryRemoveValueAt(row_index, col_index)
        return cells_collection

    def RemoveValueAt(self, row_index: int, col_index: int) -> None:
        this: FsCellsCollection = self
        get_item_from_dict(get_item_from_dict(this._rowsCollection, row_index), col_index).Value = ""

    @staticmethod
    def remove_value_at(row_index: int, col_index: int, cells_collection: FsCellsCollection) -> FsCellsCollection:
        cells_collection.RemoveValueAt(row_index, col_index)
        return cells_collection

    def GetCells(self, __unit: None=None) -> IEnumerable_1[FsCell]:
        this: FsCellsCollection = self
        def mapping(columns_collection: Any) -> Any:
            return columns_collection.values()

        return collect(mapping, this._rowsCollection.values())

    @staticmethod
    def get_cells(cells_collection: FsCellsCollection) -> IEnumerable_1[FsCell]:
        return cells_collection.GetCells()

    def GetCellsInRangeBy(self, row_start: int, column_start: int, row_end: int, column_end: int, predicate: Callable[[FsCell], bool]) -> IEnumerable_1[FsCell]:
        this: FsCellsCollection = self
        final_row: int = (this._maxRowUsed if (row_end > this._maxRowUsed) else row_end) or 0
        final_column: int = (this._maxColumnUsed if (column_end > this._maxColumnUsed) else column_end) or 0
        def _arrow206(__unit: None=None) -> IEnumerable_1[FsCell]:
            def _arrow205(ro: int) -> IEnumerable_1[FsCell]:
                match_value: Any | None = Dictionary_tryGet(ro, this._rowsCollection)
                if match_value is None:
                    return empty()

                else: 
                    columns_collection: Any = match_value
                    def _arrow204(co: int) -> IEnumerable_1[FsCell]:
                        match_value_1: FsCell | None = Dictionary_tryGet(co, columns_collection)
                        (pattern_matching_result, cell_1) = (None, None)
                        if match_value_1 is not None:
                            if predicate(match_value_1):
                                pattern_matching_result = 0
                                cell_1 = match_value_1

                            else: 
                                pattern_matching_result = 1


                        else: 
                            pattern_matching_result = 1

                        if pattern_matching_result == 0:
                            return singleton(cell_1)

                        elif pattern_matching_result == 1:
                            return empty()


                    return collect(_arrow204, range_big_int(column_start, 1, final_column))


            return collect(_arrow205, range_big_int(row_start, 1, final_row))

        return delay(_arrow206)

    @staticmethod
    def filter_cells_from_to(row_start: int, column_start: int, row_end: int, column_end: int, predicate: Callable[[FsCell], bool], cells_collection: FsCellsCollection) -> IEnumerable_1[FsCell]:
        return cells_collection.GetCellsInRangeBy(row_start, column_start, row_end, column_end, predicate)

    def GetCellsInStringRangeBy(self, start_address: FsAddress, last_address: FsAddress, predicate: Callable[[FsCell], bool]) -> IEnumerable_1[FsCell]:
        this: FsCellsCollection = self
        return this.GetCellsInRangeBy(start_address.RowNumber, start_address.ColumnNumber, last_address.RowNumber, last_address.ColumnNumber, predicate)

    @staticmethod
    def filter_cells_from_to_address(start_address: FsAddress, last_address: FsAddress, predicate: Callable[[FsCell], bool], cells_collection: FsCellsCollection) -> IEnumerable_1[FsCell]:
        return cells_collection.GetCellsInStringRangeBy(start_address, last_address, predicate)

    def GetCellsInRange(self, row_start: int, column_start: int, row_end: int, column_end: int) -> IEnumerable_1[FsCell]:
        this: FsCellsCollection = self
        final_row: int = (this._maxRowUsed if (row_end > this._maxRowUsed) else row_end) or 0
        final_column: int = (this._maxColumnUsed if (column_end > this._maxColumnUsed) else column_end) or 0
        def _arrow209(__unit: None=None) -> IEnumerable_1[FsCell]:
            def _arrow208(ro: int) -> IEnumerable_1[FsCell]:
                match_value: Any | None = Dictionary_tryGet(ro, this._rowsCollection)
                if match_value is None:
                    return empty()

                else: 
                    columns_collection: Any = match_value
                    def _arrow207(co: int) -> IEnumerable_1[FsCell]:
                        match_value_1: FsCell | None = Dictionary_tryGet(co, columns_collection)
                        if match_value_1 is not None:
                            return singleton(match_value_1)

                        else: 
                            return empty()


                    return collect(_arrow207, range_big_int(column_start, 1, final_column))


            return collect(_arrow208, range_big_int(row_start, 1, final_row))

        return delay(_arrow209)

    @staticmethod
    def get_cells_from_to(row_start: int, column_start: int, row_end: int, column_end: int, cells_collection: FsCellsCollection) -> IEnumerable_1[FsCell]:
        return cells_collection.GetCellsInRange(row_start, column_start, row_end, column_end)

    def GetCellsInStringRange(self, start_address: FsAddress, last_address: FsAddress) -> IEnumerable_1[FsCell]:
        this: FsCellsCollection = self
        return this.GetCellsInRange(start_address.RowNumber, start_address.ColumnNumber, last_address.RowNumber, last_address.ColumnNumber)

    @staticmethod
    def get_cells_from_to_address(start_address: FsAddress, last_address: FsAddress, cells_collection: FsCellsCollection) -> IEnumerable_1[FsCell]:
        return cells_collection.GetCellsInStringRange(start_address, last_address)

    def TryGetCell(self, row: int, column: int) -> FsCell | None:
        this: FsCellsCollection = self
        if True if (row > this._maxRowUsed) else (column > this._maxColumnUsed):
            return None

        else: 
            match_value: Any | None = Dictionary_tryGet(row, this._rowsCollection)
            if match_value is None:
                return None

            else: 
                match_value_1: FsCell | None = Dictionary_tryGet(column, match_value)
                return None if (match_value_1 is None) else match_value_1



    @staticmethod
    def try_get_cell(row_index: int, col_index: int, cells_collection: FsCellsCollection) -> FsCell | None:
        return cells_collection.TryGetCell(row_index, col_index)

    def GetCellsInColumn(self, col_index: int) -> IEnumerable_1[FsCell]:
        this: FsCellsCollection = self
        return this.GetCellsInRange(1, col_index, this._maxRowUsed, col_index)

    @staticmethod
    def get_cells_in_column(col_index: int, cells_collection: FsCellsCollection) -> IEnumerable_1[FsCell]:
        return cells_collection.GetCellsInColumn(col_index)

    def GetCellsInRow(self, row_index: int) -> IEnumerable_1[FsCell]:
        this: FsCellsCollection = self
        return this.GetCellsInRange(row_index, 1, row_index, this._maxColumnUsed)

    @staticmethod
    def get_cells_in_row(row_index: int, cells_collection: FsCellsCollection) -> IEnumerable_1[FsCell]:
        return cells_collection.GetCellsInRow(row_index)

    def GetFirstAddress(self, __unit: None=None) -> FsAddress:
        this: FsCellsCollection = self
        class ObjectExpr210:
            @property
            def Compare(self) -> Callable[[int, int], int]:
                return compare_primitives

        def _arrow214(__unit: None=None) -> int:
            def projection(d: Any) -> int:
                class ObjectExpr211:
                    @property
                    def Compare(self) -> Callable[[int, int], int]:
                        return compare_primitives

                return min(d.keys(), ObjectExpr211())

            class ObjectExpr212:
                @property
                def Compare(self) -> Callable[[int, int], int]:
                    return compare_primitives

            d_1: Any = min_by(projection, this._rowsCollection.values(), ObjectExpr212())
            class ObjectExpr213:
                @property
                def Compare(self) -> Callable[[int, int], int]:
                    return compare_primitives

            return min(d_1.keys(), ObjectExpr213())

        return FsAddress(0, 0) if (True if is_empty(this._rowsCollection) else is_empty(this._rowsCollection.keys())) else FsAddress(min(this._rowsCollection.keys(), ObjectExpr210()), _arrow214())

    @staticmethod
    def get_first_address(cells: FsCellsCollection) -> FsAddress:
        return cells.GetFirstAddress()

    def GetLastAddress(self, __unit: None=None) -> FsAddress:
        this: FsCellsCollection = self
        return FsAddress(this.MaxRowNumber, this.MaxColumnNumber)

    @staticmethod
    def get_last_address(cells: FsCellsCollection) -> FsAddress:
        return cells.GetLastAddress()


FsCellsCollection_reflection = _expr215

def FsCellsCollection__ctor(__unit: None=None) -> FsCellsCollection:
    return FsCellsCollection(__unit)


__all__ = ["Dictionary_tryGet", "FsCellsCollection_reflection"]




from __future__ import annotations
from collections.abc import Callable
from typing import (Any, TypeVar)
from ...arctrl_core.ontology_annotation import OntologyAnnotation
from ...arctrl_core.Table.arc_table import ArcTable
from ...arctrl_core.Table.arc_table_aux import (ColumnValueRefs, ensure_cell_hash_in_value_map, ArcTableValues, Unchecked_setCellAt)
from ...arctrl_core.Table.composite_cell import CompositeCell
from ...arctrl_core.Table.composite_header import CompositeHeader
from ...fable_library.array_ import (iterate_indexed, fold, fill)
from ...fable_library.list import (FSharpList, empty as empty_1, of_array)
from ...fable_library.map import empty as empty_2
from ...fable_library.map_util import get_item_from_dict
from ...fable_library.mutable_map import Dictionary
from ...fable_library.option import default_arg
from ...fable_library.result import FSharpResult_2
from ...fable_library.seq import (to_list, delay, append, singleton, map, empty)
from ...fable_library.types import Array
from ...fable_library.util import (IEnumerable_1, compare_arrays, equal_arrays, array_hash, get_enumerator, dispose, int32_to_string, ignore)
from ...thoth_json_core.decode import (object, list_1 as list_1_1, IOptionalGetter, map_0027, tuple2, int_1, IRequiredGetter, string, IGetters, map as map_1, resize_array, array as array_1, Helpers_prependPath)
from ...thoth_json_core.encode import list_1
from ...thoth_json_core.types import (IEncodable, IEncoderHelpers_1, Decoder_1, IDecoderHelpers_1, ErrorReason_1)
from ..decode import (Decode_tryOneOf, Decode_intDictionary, Decode_dictionary)
from ..encode import int_dictionary
from ..string_table import (encode_string, decode_string)
from .cell_table import (decode_cell, encode_cell)
from .composite_cell import (encoder as encoder_1, decoder as decoder_3)
from .composite_header import (encoder as encoder_2, decoder as decoder_2)

__A_ = TypeVar("__A_")

_VALUE_ = TypeVar("_VALUE_")

_VALUE = TypeVar("_VALUE")

def encoder(table: ArcTable) -> IEncodable:
    value_map: Any = table.Values.ValueMap
    def cell_encoder(hash_1: int, table: Any=table) -> IEncodable:
        return encoder_1(get_item_from_dict(value_map, hash_1))

    def _arrow2779(__unit: None=None, table: Any=table) -> IEnumerable_1[tuple[str, IEncodable]]:
        def _arrow2772(__unit: None=None) -> IEncodable:
            value: str = table.Name
            class ObjectExpr2771(IEncodable):
                def Encode(self, helpers: IEncoderHelpers_1[Any]) -> Any:
                    return helpers.encode_string(value)

            return ObjectExpr2771()

        def _arrow2778(__unit: None=None) -> IEnumerable_1[tuple[str, IEncodable]]:
            def _arrow2773(__unit: None=None) -> IEnumerable_1[IEncodable]:
                return map(encoder_2, table.Headers)

            def _arrow2777(__unit: None=None) -> IEnumerable_1[tuple[str, IEncodable]]:
                def column_encoder(col: ColumnValueRefs) -> IEncodable:
                    if col.tag == 1:
                        return int_dictionary(cell_encoder, col.fields[0])

                    else: 
                        return cell_encoder(col.fields[0])


                def _arrow2776(__unit: None=None) -> IEnumerable_1[tuple[str, IEncodable]]:
                    def _arrow2775(__unit: None=None) -> IEncodable:
                        value_1: int = table.RowCount or 0
                        class ObjectExpr2774(IEncodable):
                            def Encode(self, helpers_1: IEncoderHelpers_1[Any]) -> Any:
                                return helpers_1.encode_signed_integral_number(value_1)

                        return ObjectExpr2774()

                    return singleton(("rowCount", _arrow2775()))

                return append(singleton(("columns", int_dictionary(column_encoder, table.Values.Columns))), delay(_arrow2776)) if (table.Values.RowCount != 0) else empty()

            return append(singleton(("headers", list_1(to_list(delay(_arrow2773))))) if (len(table.Headers) != 0) else empty(), delay(_arrow2777))

        return append(singleton(("name", _arrow2772())), delay(_arrow2778))

    values: IEnumerable_1[tuple[str, IEncodable]] = to_list(delay(_arrow2779))
    class ObjectExpr2780(IEncodable):
        def Encode(self, helpers_2: IEncoderHelpers_1[Any], table: Any=table) -> Any:
            def mapping(tupled_arg: tuple[str, IEncodable]) -> tuple[str, __A_]:
                return (tupled_arg[0], tupled_arg[1].Encode(helpers_2))

            arg: IEnumerable_1[tuple[str, __A_]] = map(mapping, values)
            return helpers_2.encode_object(arg)

    return ObjectExpr2780()


def _arrow2786(get: IGetters) -> ArcTable:
    def _arrow2781(__unit: None=None) -> FSharpList[CompositeHeader] | None:
        arg_1: Decoder_1[FSharpList[CompositeHeader]] = list_1_1(decoder_2)
        object_arg: IOptionalGetter = get.Optional
        return object_arg.Field("header", arg_1)

    decoded_header: Array[CompositeHeader] = list(default_arg(_arrow2781(), empty_1()))
    def _arrow2782(__unit: None=None) -> Any | None:
        arg_3: Decoder_1[Any] = map_0027(tuple2(int_1, int_1), decoder_3)
        object_arg_1: IOptionalGetter = get.Optional
        return object_arg_1.Field("values", arg_3)

    class ObjectExpr2783:
        @property
        def Compare(self) -> Callable[[tuple[int, int], tuple[int, int]], int]:
            return compare_arrays

    class ObjectExpr2784:
        @property
        def Equals(self) -> Callable[[tuple[int, int], tuple[int, int]], bool]:
            return equal_arrays

        @property
        def GetHashCode(self) -> Callable[[tuple[int, int]], int]:
            return array_hash

    decoded_values: Any = Dictionary(default_arg(_arrow2782(), empty_2(ObjectExpr2783())), ObjectExpr2784())
    def _arrow2785(__unit: None=None) -> str:
        object_arg_2: IRequiredGetter = get.Required
        return object_arg_2.Field("name", string)

    t: ArcTable = ArcTable.create(_arrow2785(), decoded_header, [])
    enumerator: Any = get_enumerator(decoded_values)
    try: 
        while enumerator.System_Collections_IEnumerator_MoveNext():
            active_pattern_result: tuple[tuple[int, int], CompositeCell] = enumerator.System_Collections_Generic_IEnumerator_1_get_Current()
            t.SetCellAt(active_pattern_result[0][0], active_pattern_result[0][1], active_pattern_result[1])

    finally: 
        dispose(enumerator)

    return t


decoder_v2deprecated: Decoder_1[ArcTable] = object(_arrow2786)

def _arrow2793(__unit: None=None) -> Decoder_1[ArcTable]:
    value_map: Any = dict([])
    def ctor(cell: CompositeCell) -> int:
        return ensure_cell_hash_in_value_map(cell, value_map)

    cell_decoder: Decoder_1[int] = map_1(ctor, decoder_3)
    def ctor_1(Item: int) -> ColumnValueRefs:
        return ColumnValueRefs(0, Item)

    def ctor_2(Item_1: Any) -> ColumnValueRefs:
        return ColumnValueRefs(1, Item_1)

    column_decoder: Decoder_1[ColumnValueRefs] = Decode_tryOneOf(of_array([map_1(ctor_1, cell_decoder), map_1(ctor_2, Decode_intDictionary(cell_decoder))]))
    def _arrow2791(get: IGetters) -> ArcTable:
        def _arrow2787(__unit: None=None) -> Array[CompositeHeader] | None:
            arg_1: Decoder_1[Array[CompositeHeader]] = resize_array(decoder_2)
            object_arg: IOptionalGetter = get.Optional
            return object_arg.Field("headers", arg_1)

        decoded_header: Array[CompositeHeader] = default_arg(_arrow2787(), [])
        def _arrow2788(__unit: None=None) -> Any | None:
            arg_3: Decoder_1[Any] = Decode_dictionary(int_1, column_decoder)
            object_arg_1: IOptionalGetter = get.Optional
            return object_arg_1.Field("columns", arg_3)

        def _arrow2789(__unit: None=None) -> int | None:
            object_arg_2: IOptionalGetter = get.Optional
            return object_arg_2.Field("rowCount", int_1)

        values: ArcTableValues = ArcTableValues(default_arg(_arrow2788(), dict([])), value_map, default_arg(_arrow2789(), 0))
        def _arrow2790(__unit: None=None) -> str:
            object_arg_3: IRequiredGetter = get.Required
            return object_arg_3.Field("name", string)

        return ArcTable.from_arc_table_values(_arrow2790(), decoded_header, values)

    decoder_1: Decoder_1[ArcTable] = object(_arrow2791)
    class ObjectExpr2792(Decoder_1[ArcTable]):
        def Decode(self, helper: IDecoderHelpers_1[Any], column: Any) -> FSharpResult_2[ArcTable, tuple[str, ErrorReason_1[__A_]]]:
            return decoder_v2deprecated.Decode(helper, column) if helper.has_property("values", column) else decoder_1.Decode(helper, column)

    return ObjectExpr2792()


decoder: Decoder_1[ArcTable] = _arrow2793()

def decoder_compressed_column(cell_table: Array[CompositeCell], table: ArcTable, column_index: int) -> Decoder_1[None]:
    class ObjectExpr2795(Decoder_1[None]):
        def Decode(self, helper: IDecoderHelpers_1[Any], column: Any, cell_table: Any=cell_table, table: Any=table, column_index: Any=column_index) -> FSharpResult_2[None, tuple[str, ErrorReason_1[__A_]]]:
            match_value: FSharpResult_2[Array[CompositeCell], tuple[str, ErrorReason_1[__A_]]] = array_1(decode_cell(cell_table)).Decode(helper, column)
            if match_value.tag == 1:
                def _arrow2794(get: IGetters) -> None:
                    from_: int
                    object_arg: IRequiredGetter = get.Required
                    from_ = object_arg.Field("f", int_1)
                    to_: int
                    object_arg_1: IRequiredGetter = get.Required
                    to_ = object_arg_1.Field("t", int_1)
                    value: CompositeCell
                    arg_5: Decoder_1[CompositeCell] = decode_cell(cell_table)
                    object_arg_2: IRequiredGetter = get.Required
                    value = object_arg_2.Field("v", arg_5)
                    for i in range(from_, to_ + 1, 1):
                        Unchecked_setCellAt(column_index, i, value, table.Values)

                range_decoder: Decoder_1[None] = object(_arrow2794)
                match_value_1: FSharpResult_2[Array[None], tuple[str, ErrorReason_1[__A_]]] = array_1(range_decoder).Decode(helper, column)
                return FSharpResult_2(1, match_value_1.fields[0]) if (match_value_1.tag == 1) else FSharpResult_2(0, None)

            else: 
                def action(r: int, cell: CompositeCell) -> None:
                    Unchecked_setCellAt(column_index, r, cell, table.Values)

                iterate_indexed(action, match_value.fields[0])
                return FSharpResult_2(0, None)


    return ObjectExpr2795()


def arrayi(decoderi: Callable[[int], Decoder_1[_VALUE]]) -> Decoder_1[Array[Any]]:
    class ObjectExpr2797(Decoder_1[Array[_VALUE_]]):
        def Decode(self, helpers: IDecoderHelpers_1[Any], value: Any, decoderi: Any=decoderi) -> FSharpResult_2[Array[_VALUE_], tuple[str, ErrorReason_1[__A_]]]:
            if helpers.is_array(value):
                i: int = -1
                tokens: Array[__A_] = helpers.as_array(value)
                def folder(acc: FSharpResult_2[Array[_VALUE_], tuple[str, ErrorReason_1[__A_]]], value_1: __A_) -> FSharpResult_2[Array[_VALUE_], tuple[str, ErrorReason_1[__A_]]]:
                    nonlocal i
                    i = (i + 1) or 0
                    if acc.tag == 0:
                        acc_1: Array[_VALUE_] = acc.fields[0]
                        match_value: FSharpResult_2[_VALUE_, tuple[str, ErrorReason_1[__A_]]] = decoderi(i).Decode(helpers, value_1)
                        if match_value.tag == 0:
                            acc_1[i] = match_value.fields[0]
                            return FSharpResult_2(0, acc_1)

                        else: 
                            def _arrow2796(__unit: None=None, acc: Any=acc, value_1: Any=value_1) -> tuple[str, ErrorReason_1[__A_]]:
                                tupled_arg: tuple[str, ErrorReason_1[__A_]] = match_value.fields[0]
                                return Helpers_prependPath((".[" + int32_to_string(i)) + "]", tupled_arg[0], tupled_arg[1])

                            return FSharpResult_2(1, _arrow2796())


                    else: 
                        return acc


                return fold(folder, FSharpResult_2(0, fill([0] * len(tokens), 0, len(tokens), None)), tokens)

            else: 
                return FSharpResult_2(1, ("", ErrorReason_1(0, "an array", value)))


    return ObjectExpr2797()


def encoder_compressed(string_table: Any, oa_table: Any, cell_table: Any, table: ArcTable) -> IEncodable:
    def cell_encoder(hash_1: int, string_table: Any=string_table, oa_table: Any=oa_table, cell_table: Any=cell_table, table: Any=table) -> IEncodable:
        return encode_cell(cell_table, get_item_from_dict(table.Values.ValueMap, hash_1))

    def _arrow2804(__unit: None=None, string_table: Any=string_table, oa_table: Any=oa_table, cell_table: Any=cell_table, table: Any=table) -> IEnumerable_1[tuple[str, IEncodable]]:
        def _arrow2803(__unit: None=None) -> IEnumerable_1[tuple[str, IEncodable]]:
            def _arrow2798(__unit: None=None) -> IEnumerable_1[IEncodable]:
                return map(encoder_2, table.Headers)

            def _arrow2802(__unit: None=None) -> IEnumerable_1[tuple[str, IEncodable]]:
                def column_encoder(col: ColumnValueRefs) -> IEncodable:
                    if col.tag == 1:
                        return int_dictionary(cell_encoder, col.fields[0])

                    else: 
                        return cell_encoder(col.fields[0])


                def _arrow2801(__unit: None=None) -> IEnumerable_1[tuple[str, IEncodable]]:
                    def _arrow2800(__unit: None=None) -> IEncodable:
                        value: int = table.RowCount or 0
                        class ObjectExpr2799(IEncodable):
                            def Encode(self, helpers: IEncoderHelpers_1[Any]) -> Any:
                                return helpers.encode_signed_integral_number(value)

                        return ObjectExpr2799()

                    return singleton(("r", _arrow2800()))

                return append(singleton(("c", int_dictionary(column_encoder, table.Values.Columns))), delay(_arrow2801)) if (table.Values.RowCount != 0) else empty()

            return append(singleton(("h", list_1(to_list(delay(_arrow2798))))) if (len(table.Headers) != 0) else empty(), delay(_arrow2802))

        return append(singleton(("n", encode_string(string_table, table.Name))), delay(_arrow2803))

    values: IEnumerable_1[tuple[str, IEncodable]] = to_list(delay(_arrow2804))
    class ObjectExpr2805(IEncodable):
        def Encode(self, helpers_1: IEncoderHelpers_1[Any], string_table: Any=string_table, oa_table: Any=oa_table, cell_table: Any=cell_table, table: Any=table) -> Any:
            def mapping(tupled_arg: tuple[str, IEncodable]) -> tuple[str, __A_]:
                return (tupled_arg[0], tupled_arg[1].Encode(helpers_1))

            arg: IEnumerable_1[tuple[str, __A_]] = map(mapping, values)
            return helpers_1.encode_object(arg)

    return ObjectExpr2805()


def decoder_compressed_v2deprecated(string_table: Array[str], oa_table: Array[OntologyAnnotation], cell_table: Array[CompositeCell]) -> Decoder_1[ArcTable]:
    def _arrow2812(get: IGetters, string_table: Any=string_table, oa_table: Any=oa_table, cell_table: Any=cell_table) -> ArcTable:
        def _arrow2806(__unit: None=None) -> FSharpList[CompositeHeader] | None:
            arg_1: Decoder_1[FSharpList[CompositeHeader]] = list_1_1(decoder_2)
            object_arg: IOptionalGetter = get.Optional
            return object_arg.Field("h", arg_1)

        decoded_header: Array[CompositeHeader] = list(default_arg(_arrow2806(), empty_1()))
        def _arrow2809(__unit: None=None) -> str:
            arg_3: Decoder_1[str] = decode_string(string_table)
            object_arg_1: IRequiredGetter = get.Required
            return object_arg_1.Field("n", arg_3)

        table: ArcTable = ArcTable.create(_arrow2809(), decoded_header, [])
        def _arrow2811(__unit: None=None) -> Array[None] | None:
            def _arrow2810(column_index: int) -> Decoder_1[None]:
                return decoder_compressed_column(cell_table, table, column_index)

            arg_5: Decoder_1[Array[None]] = arrayi(_arrow2810)
            object_arg_2: IOptionalGetter = get.Optional
            return object_arg_2.Field("c", arg_5)

        ignore(_arrow2811())
        return table

    return object(_arrow2812)


def decoder_compressed(string_table: Array[str], oa_table: Array[OntologyAnnotation], cell_table: Array[CompositeCell]) -> Decoder_1[ArcTable]:
    value_map: Any = dict([])
    def ctor(i: int, string_table: Any=string_table, oa_table: Any=oa_table, cell_table: Any=cell_table) -> int:
        return ensure_cell_hash_in_value_map(cell_table[i], value_map)

    cell_decoder: Decoder_1[int] = map_1(ctor, int_1)
    def ctor_1(Item: int, string_table: Any=string_table, oa_table: Any=oa_table, cell_table: Any=cell_table) -> ColumnValueRefs:
        return ColumnValueRefs(0, Item)

    def ctor_2(Item_1: Any, string_table: Any=string_table, oa_table: Any=oa_table, cell_table: Any=cell_table) -> ColumnValueRefs:
        return ColumnValueRefs(1, Item_1)

    column_decoder: Decoder_1[ColumnValueRefs] = Decode_tryOneOf(of_array([map_1(ctor_1, cell_decoder), map_1(ctor_2, Decode_intDictionary(cell_decoder))]))
    def _arrow2832(get: IGetters, string_table: Any=string_table, oa_table: Any=oa_table, cell_table: Any=cell_table) -> ArcTable:
        def _arrow2818(__unit: None=None) -> Array[CompositeHeader] | None:
            arg_1: Decoder_1[Array[CompositeHeader]] = resize_array(decoder_2)
            object_arg: IOptionalGetter = get.Optional
            return object_arg.Field("h", arg_1)

        decoded_header: Array[CompositeHeader] = default_arg(_arrow2818(), [])
        def _arrow2825(__unit: None=None) -> Any | None:
            arg_3: Decoder_1[Any] = Decode_dictionary(int_1, column_decoder)
            object_arg_1: IOptionalGetter = get.Optional
            return object_arg_1.Field("c", arg_3)

        def _arrow2826(__unit: None=None) -> int | None:
            object_arg_2: IOptionalGetter = get.Optional
            return object_arg_2.Field("r", int_1)

        values: ArcTableValues = ArcTableValues(default_arg(_arrow2825(), dict([])), value_map, default_arg(_arrow2826(), 0))
        def _arrow2831(__unit: None=None) -> str:
            arg_7: Decoder_1[str] = decode_string(string_table)
            object_arg_3: IRequiredGetter = get.Required
            return object_arg_3.Field("n", arg_7)

        return ArcTable.from_arc_table_values(_arrow2831(), decoded_header, values)

    decoder_1: Decoder_1[ArcTable] = object(_arrow2832)
    class ObjectExpr2836(Decoder_1[ArcTable]):
        def Decode(self, helper: IDecoderHelpers_1[Any], column: Any, string_table: Any=string_table, oa_table: Any=oa_table, cell_table: Any=cell_table) -> FSharpResult_2[ArcTable, tuple[str, ErrorReason_1[__A_]]]:
            return decoder_compressed_v2deprecated(string_table, oa_table, cell_table).Decode(helper, column) if (not helper.has_property("r", column)) else decoder_1.Decode(helper, column)

    return ObjectExpr2836()


__all__ = ["encoder", "decoder_v2deprecated", "decoder", "decoder_compressed_column", "arrayi", "encoder_compressed", "decoder_compressed_v2deprecated", "decoder_compressed"]




from __future__ import annotations
from typing import Any
from ...dynamic_obj.dyn_obj import set_optional_property
from ...fable_library.option import default_arg
from ...fable_library.reflection import (TypeInfo, class_type)
from ...fable_library.types import (FSharpRef, Array)
from ..ldobject import (LDNode, LDNode_reflection)

def _expr1554() -> TypeInfo:
    return class_type("ARCtrl.ROCrate.LabProtocol", None, LabProtocol, LDNode_reflection())


class LabProtocol(LDNode):
    def __init__(self, id: str, additional_type: Array[str] | None=None, name: Any | None=None, intended_use: Any | None=None, description: Any | None=None, url: Any | None=None, comment: Any | None=None, version: Any | None=None, lab_equipment: Any | None=None, reagent: Any | None=None, computational_tool: Any | None=None) -> None:
        super().__init__(id, ["bioschemas.org/LabProtocol"], default_arg(additional_type, []))
        this: FSharpRef[LabProtocol] = FSharpRef(None)
        this.contents = self
        self.init_00408: int = 1
        set_optional_property("name", name, this.contents)
        set_optional_property("intendedUse", intended_use, this.contents)
        set_optional_property("description", description, this.contents)
        set_optional_property("url", url, this.contents)
        set_optional_property("comment", comment, this.contents)
        set_optional_property("version", version, this.contents)
        set_optional_property("labEquipment", lab_equipment, this.contents)
        set_optional_property("reagent", reagent, this.contents)
        set_optional_property("computationalTool", computational_tool, this.contents)


LabProtocol_reflection = _expr1554

def LabProtocol__ctor_Z6F48851E(id: str, additional_type: Array[str] | None=None, name: Any | None=None, intended_use: Any | None=None, description: Any | None=None, url: Any | None=None, comment: Any | None=None, version: Any | None=None, lab_equipment: Any | None=None, reagent: Any | None=None, computational_tool: Any | None=None) -> LabProtocol:
    return LabProtocol(id, additional_type, name, intended_use, description, url, comment, version, lab_equipment, reagent, computational_tool)


__all__ = ["LabProtocol_reflection"]




from __future__ import annotations
from datetime import datetime
from typing import Any
from ...arctrl_core.Helper.collections_ import ResizeArray_map
from ...arctrl_core.Helper.identifier import create_missing_identifier
from ...fable_library.option import (value, default_arg)
from ...fable_library.reflection import (TypeInfo, class_type)
from ...fable_library.types import Array
from ..helper import clean
from ..ldcontext import LDContext
from ..ldobject import (LDNode, LDGraph)
from .computational_workflow import LDComputationalWorkflow
from .file import LDFile
from .person import LDPerson
from .sample import LDSample

def _expr1536() -> TypeInfo:
    return class_type("ARCtrl.ROCrate.LDCreateAction", None, LDCreateAction)


class LDCreateAction:
    @staticmethod
    def schema_type() -> str:
        return "http://schema.org/CreateAction"

    @staticmethod
    def name() -> str:
        return "http://schema.org/name"

    @staticmethod
    def instrument() -> str:
        return "http://schema.org/instrument"

    @staticmethod
    def agent() -> str:
        return "http://schema.org/agent"

    @staticmethod
    def object_() -> str:
        return "http://schema.org/object"

    @staticmethod
    def result() -> str:
        return "http://schema.org/result"

    @staticmethod
    def description() -> str:
        return "http://schema.org/description"

    @staticmethod
    def end_time() -> str:
        return "http://schema.org/endTime"

    @staticmethod
    def disambiguating_description() -> str:
        return "http://schema.org/disambiguatingDescription"

    @staticmethod
    def try_get_name_as_string(ca: LDNode, context: LDContext | None=None) -> str | None:
        match_value: Any | None = ca.TryGetPropertyAsSingleton(LDCreateAction.name(), context)
        (pattern_matching_result, n) = (None, None)
        if match_value is not None:
            if str(type(value(match_value))) == "<class \'str\'>":
                pattern_matching_result = 0
                n = value(match_value)

            else: 
                pattern_matching_result = 1


        else: 
            pattern_matching_result = 1

        if pattern_matching_result == 0:
            return n

        elif pattern_matching_result == 1:
            return None


    @staticmethod
    def get_name_as_string(ca: LDNode, context: LDContext | None=None) -> str:
        match_value: Any | None = ca.TryGetPropertyAsSingleton(LDCreateAction.name(), context)
        if match_value is not None:
            if str(type(value(match_value))) == "<class \'str\'>":
                n: str = value(match_value)
                return n

            else: 
                raise Exception(("property `name` of object with @id `" + ca.Id) + "` was not a string")


        else: 
            raise Exception(("Could not access property `name` of object with @id `" + ca.Id) + "`")


    @staticmethod
    def set_name_as_string(ca: LDNode, name: str, context: LDContext | None=None) -> Any:
        return ca.SetProperty(LDCreateAction.name(), name, context)

    @staticmethod
    def get_instruments(ca: LDNode, graph: LDGraph | None=None, context: LDContext | None=None) -> Array[LDNode]:
        return ca.GetPropertyNodes(LDCreateAction.instrument(), None, graph, context)

    @staticmethod
    def get_instruments_as_computational_workflow(ca: LDNode, graph: LDGraph | None=None, context: LDContext | None=None) -> Array[LDNode]:
        def filter(ld_object: LDNode, context_1: LDContext | None=None) -> bool:
            return LDComputationalWorkflow.validate(ld_object, context_1)

        return ca.GetPropertyNodes(LDCreateAction.instrument(), filter, graph, context)

    @staticmethod
    def set_instruments(ca: LDNode, instruments: Array[LDNode], context: LDContext | None=None) -> Any:
        return ca.SetProperty(LDCreateAction.instrument(), instruments, context)

    @staticmethod
    def try_get_agent(ca: LDNode, graph: LDGraph | None=None, context: LDContext | None=None) -> LDNode | None:
        match_value: LDNode | None = ca.TryGetPropertyAsSingleNode(LDCreateAction.agent(), graph, context)
        (pattern_matching_result, a_1) = (None, None)
        if match_value is not None:
            def _arrow1534(__unit: None=None) -> bool:
                ld_object: LDNode = match_value
                return LDPerson.validate(ld_object, context)

            if _arrow1534():
                pattern_matching_result = 0
                a_1 = match_value

            else: 
                pattern_matching_result = 1


        else: 
            pattern_matching_result = 1

        if pattern_matching_result == 0:
            return a_1

        elif pattern_matching_result == 1:
            return None


    @staticmethod
    def get_agent(ca: LDNode, graph: LDGraph | None=None, context: LDContext | None=None) -> LDNode:
        match_value: LDNode | None = ca.TryGetPropertyAsSingleNode(LDCreateAction.agent(), graph, context)
        if match_value is not None:
            def _arrow1535(__unit: None=None) -> bool:
                ld_object: LDNode = match_value
                return LDPerson.validate(ld_object, context)

            if _arrow1535():
                a_1: LDNode = match_value
                return a_1

            else: 
                raise Exception(("Property of `agent` of object with @id `" + ca.Id) + "` was not a valid Person")


        else: 
            raise Exception(("Could not access property `agent` of object with @id `" + ca.Id) + "`")


    @staticmethod
    def set_agent(ca: LDNode, agent: LDNode, context: LDContext | None=None) -> Any:
        return ca.SetProperty(LDCreateAction.agent(), agent, context)

    @staticmethod
    def get_objects(ca: LDNode, graph: LDGraph | None=None, context: LDContext | None=None) -> Array[LDNode]:
        return ca.GetPropertyNodes(LDCreateAction.object_(), None, graph, context)

    @staticmethod
    def get_objects_as_sample(ca: LDNode, graph: LDGraph | None=None, context: LDContext | None=None) -> Array[LDNode]:
        def filter(ld_object: LDNode, context_1: LDContext | None=None) -> bool:
            return LDSample.validate(ld_object, context_1)

        return ca.GetPropertyNodes(LDCreateAction.object_(), filter, graph, context)

    @staticmethod
    def get_objects_as_data(ca: LDNode, graph: LDGraph | None=None, context: LDContext | None=None) -> Array[LDNode]:
        def filter(ld_object: LDNode, context_1: LDContext | None=None) -> bool:
            return LDFile.validate(ld_object, context_1)

        return ca.GetPropertyNodes(LDCreateAction.object_(), filter, graph, context)

    @staticmethod
    def set_objects(ca: LDNode, objects: Array[LDNode], context: LDContext | None=None) -> Any:
        return ca.SetProperty(LDCreateAction.object_(), objects, context)

    @staticmethod
    def get_results(ca: LDNode, graph: LDGraph | None=None, context: LDContext | None=None) -> Array[LDNode]:
        return ca.GetPropertyNodes(LDCreateAction.result(), None, graph, context)

    @staticmethod
    def get_results_as_sample(ca: LDNode, graph: LDGraph | None=None, context: LDContext | None=None) -> Array[LDNode]:
        def filter(ld_object: LDNode, context_1: LDContext | None=None) -> bool:
            return LDSample.validate(ld_object, context_1)

        return ca.GetPropertyNodes(LDCreateAction.result(), filter, graph, context)

    @staticmethod
    def get_results_as_data(ca: LDNode, graph: LDGraph | None=None, context: LDContext | None=None) -> Array[LDNode]:
        def filter(ld_object: LDNode, context_1: LDContext | None=None) -> bool:
            return LDFile.validate(ld_object, context_1)

        return ca.GetPropertyNodes(LDCreateAction.result(), filter, graph, context)

    @staticmethod
    def set_results(ca: LDNode, results: Array[LDNode], context: LDContext | None=None) -> Any:
        return ca.SetProperty(LDCreateAction.result(), results, context)

    @staticmethod
    def try_get_description(ca: LDNode, context: LDContext | None=None) -> str | None:
        match_value: Any | None = ca.TryGetPropertyAsSingleton(LDCreateAction.description(), context)
        (pattern_matching_result, d) = (None, None)
        if match_value is not None:
            if str(type(value(match_value))) == "<class \'str\'>":
                pattern_matching_result = 0
                d = value(match_value)

            else: 
                pattern_matching_result = 1


        else: 
            pattern_matching_result = 1

        if pattern_matching_result == 0:
            return d

        elif pattern_matching_result == 1:
            return None


    @staticmethod
    def set_description(ca: LDNode, description: str, context: LDContext | None=None) -> Any:
        return ca.SetProperty(LDCreateAction.description(), description, context)

    @staticmethod
    def try_get_end_time(ca: LDNode, context: LDContext | None=None) -> Any | None:
        match_value: Any | None = ca.TryGetPropertyAsSingleton(LDCreateAction.end_time(), context)
        (pattern_matching_result, et) = (None, None)
        if match_value is not None:
            if isinstance(value(match_value), datetime):
                pattern_matching_result = 0
                et = value(match_value)

            else: 
                pattern_matching_result = 1


        else: 
            pattern_matching_result = 1

        if pattern_matching_result == 0:
            return et

        elif pattern_matching_result == 1:
            return None


    @staticmethod
    def set_end_time(ca: LDNode, end_time: Any, context: LDContext | None=None) -> Any:
        return ca.SetProperty(LDCreateAction.end_time(), end_time, context)

    @staticmethod
    def get_disambiguating_descriptions_as_string(ca: LDNode, context: LDContext | None=None) -> Array[str]:
        def f(o_1: Any=None) -> Any:
            return o_1

        def filter(o: Any=None, context_1: LDContext | None=None) -> bool:
            return str(type(o)) == "<class \'str\'>"

        return ResizeArray_map(f, ca.GetPropertyValues(LDCreateAction.disambiguating_description(), filter, context))

    @staticmethod
    def set_disambiguating_descriptions_as_string(ca: LDNode, disambiguating_descriptions: Array[str], context: LDContext | None=None) -> Any:
        return ca.SetProperty(LDCreateAction.disambiguating_description(), disambiguating_descriptions, context)

    @staticmethod
    def validate(ca: LDNode, context: LDContext | None=None) -> bool:
        return ca.HasProperty(LDCreateAction.instrument(), context) if ca.HasType(LDCreateAction.schema_type(), context) else False

    @staticmethod
    def create(name: str, agent: LDNode, instrument: LDNode, objects: Array[LDNode] | None=None, results: Array[LDNode] | None=None, description: str | None=None, id: str | None=None, end_time: Any | None=None, disambiguating_descriptions: Array[str] | None=None, context: LDContext | None=None) -> LDNode:
        id_2: str = clean(("#ComputationalWorkflow_" + create_missing_identifier()) + "") if (id is None) else id
        objects_1: Array[LDNode] = default_arg(objects, [])
        results_1: Array[LDNode] = default_arg(results, [])
        ca: LDNode = LDNode(id_2, [LDCreateAction.schema_type()], None, context)
        ca.SetProperty(LDCreateAction.name(), name, context)
        ca.SetProperty(LDCreateAction.agent(), agent, context)
        ca.SetProperty(LDCreateAction.object_(), objects_1, context)
        ca.SetProperty(LDCreateAction.result(), results_1, context)
        ca.SetProperty(LDCreateAction.instrument(), instrument, context)
        ca.SetOptionalProperty(LDCreateAction.description(), description, context)
        ca.SetOptionalProperty(LDCreateAction.end_time(), end_time, context)
        ca.SetOptionalProperty(LDCreateAction.disambiguating_description(), disambiguating_descriptions, context)
        return ca


LDCreateAction_reflection = _expr1536

__all__ = ["LDCreateAction_reflection"]




from __future__ import annotations
from dataclasses import dataclass
from typing import Any
from ..dynamic_obj.dynamic_obj import (DynamicObj, DynamicObj_reflection)
from ..dynamic_obj.dyn_obj import set_optional_property
from ..fable_library.option import value
from ..fable_library.reflection import (TypeInfo, string_type, option_type, int32_type, bool_type, record_type, class_type)
from ..fable_library.types import (Record, FSharpRef)
from .cwltypes import CWLType

def _expr488() -> TypeInfo:
    return record_type("ARCtrl.CWL.InputBinding", [], InputBinding, lambda: [("Prefix", option_type(string_type)), ("Position", option_type(int32_type)), ("ItemSeparator", option_type(string_type)), ("Separate", option_type(bool_type))])


@dataclass(eq = False, repr = False, slots = True)
class InputBinding(Record):
    Prefix: str | None
    Position: int | None
    ItemSeparator: str | None
    Separate: bool | None

InputBinding_reflection = _expr488

def InputBinding_create_ZAC0108A(prefix: str | None=None, position: int | None=None, item_separator: str | None=None, separate: bool | None=None) -> InputBinding:
    return InputBinding(prefix, position, item_separator, separate)


def _expr489() -> TypeInfo:
    return class_type("ARCtrl.CWL.CWLInput", None, CWLInput, DynamicObj_reflection())


class CWLInput(DynamicObj):
    def __init__(self, name: str, type_: CWLType | None=None, input_binding: InputBinding | None=None, optional: bool | None=None) -> None:
        super().__init__()
        this: FSharpRef[CWLInput] = FSharpRef(None)
        self.name: str = name
        this.contents = self
        self.init_004031: int = 1
        set_optional_property("type", type_, this.contents)
        set_optional_property("inputBinding", input_binding, this.contents)
        set_optional_property("optional", optional, this.contents)

    @property
    def Name(self, __unit: None=None) -> str:
        this: CWLInput = self
        return this.name

    @property
    def Type_(self, __unit: None=None) -> CWLType | None:
        this: CWLInput = self
        match_value: Any | None = this.TryGetPropertyValue("type")
        if match_value is not None:
            o: Any = value(match_value)
            return o if isinstance(o, CWLType) else None

        else: 
            return None


    @property
    def InputBinding(self, __unit: None=None) -> InputBinding | None:
        this: CWLInput = self
        match_value: Any | None = this.TryGetPropertyValue("inputBinding")
        if match_value is not None:
            o: Any = value(match_value)
            return o if isinstance(o, InputBinding) else None

        else: 
            return None


    @property
    def Optional(self, __unit: None=None) -> bool | None:
        this: CWLInput = self
        match_value: Any | None = this.TryGetPropertyValue("optional")
        if match_value is not None:
            o: Any = value(match_value)
            return o if (str(type(o)) == "<class \'bool\'>") else None

        else: 
            return None



CWLInput_reflection = _expr489

def CWLInput__ctor_Z3A15BEDB(name: str, type_: CWLType | None=None, input_binding: InputBinding | None=None, optional: bool | None=None) -> CWLInput:
    return CWLInput(name, type_, input_binding, optional)


__all__ = ["InputBinding_reflection", "InputBinding_create_ZAC0108A", "CWLInput_reflection"]




from __future__ import annotations
from collections.abc import Callable
from typing import (Any, TypeVar)
from ...arctrl_core.Process.material import Material
from ...arctrl_core.Process.material_attribute_value import MaterialAttributeValue
from ...arctrl_core.Process.material_type import MaterialType
from ...fable_library.list import (choose, singleton, of_array, FSharpList)
from ...fable_library.option import map
from ...fable_library.seq import map as map_1
from ...fable_library.string_ import replace
from ...fable_library.util import IEnumerable_1
from ...thoth_json_core.decode import (object, IOptionalGetter, IRequiredGetter, unit, string, list_1 as list_1_2, IGetters)
from ...thoth_json_core.encode import list_1 as list_1_1
from ...thoth_json_core.types import (IEncodable, IEncoderHelpers_1, Decoder_1)
from ..context.rocrate.isa_material_context import context_jsonvalue
from ..decode import (Decode_uri, Decode_objectNoAdditionalProperties)
from ..encode import (try_include, try_include_list_opt)
from ..idtable import encode
from .material_attribute_value import (ROCrate_encoder as ROCrate_encoder_2, ROCrate_decoder as ROCrate_decoder_2, ISAJson_encoder as ISAJson_encoder_2, ISAJson_decoder as ISAJson_decoder_2)
from .material_type import (ROCrate_encoder as ROCrate_encoder_1, ROCrate_decoder as ROCrate_decoder_1, ISAJson_encoder as ISAJson_encoder_1, ISAJson_decoder as ISAJson_decoder_1)

__A_ = TypeVar("__A_")

def ROCrate_genID(m: Material) -> str:
    match_value: str | None = m.ID
    if match_value is None:
        match_value_1: str | None = m.Name
        if match_value_1 is None:
            return "#EmptyMaterial"

        else: 
            return "#Material_" + replace(match_value_1, " ", "_")


    else: 
        return match_value



def ROCrate_encoder(oa: Material) -> IEncodable:
    def chooser(tupled_arg: tuple[str, IEncodable | None], oa: Any=oa) -> tuple[str, IEncodable] | None:
        def mapping(v_1: IEncodable, tupled_arg: Any=tupled_arg) -> tuple[str, IEncodable]:
            return (tupled_arg[0], v_1)

        return map(mapping, tupled_arg[1])

    def _arrow2552(__unit: None=None, oa: Any=oa) -> IEncodable:
        value: str = ROCrate_genID(oa)
        class ObjectExpr2551(IEncodable):
            def Encode(self, helpers: IEncoderHelpers_1[Any]) -> Any:
                return helpers.encode_string(value)

        return ObjectExpr2551()

    class ObjectExpr2553(IEncodable):
        def Encode(self, helpers_1: IEncoderHelpers_1[Any], oa: Any=oa) -> Any:
            return helpers_1.encode_string("Material")

    class ObjectExpr2554(IEncodable):
        def Encode(self, helpers_2: IEncoderHelpers_1[Any], oa: Any=oa) -> Any:
            return helpers_2.encode_string("Material")

    def _arrow2556(value_3: str, oa: Any=oa) -> IEncodable:
        class ObjectExpr2555(IEncodable):
            def Encode(self, helpers_3: IEncoderHelpers_1[Any]) -> Any:
                return helpers_3.encode_string(value_3)

        return ObjectExpr2555()

    def _arrow2557(value_5: MaterialType, oa: Any=oa) -> IEncodable:
        return ROCrate_encoder_1(value_5)

    def _arrow2558(oa_1: Material, oa: Any=oa) -> IEncodable:
        return ROCrate_encoder(oa_1)

    values: FSharpList[tuple[str, IEncodable]] = choose(chooser, of_array([("@id", _arrow2552()), ("@type", list_1_1(singleton(ObjectExpr2553()))), ("additionalType", ObjectExpr2554()), try_include("name", _arrow2556, oa.Name), try_include("type", _arrow2557, oa.MaterialType), try_include_list_opt("characteristics", ROCrate_encoder_2, oa.Characteristics), try_include_list_opt("derivesFrom", _arrow2558, oa.DerivesFrom), ("@context", context_jsonvalue)]))
    class ObjectExpr2559(IEncodable):
        def Encode(self, helpers_4: IEncoderHelpers_1[Any], oa: Any=oa) -> Any:
            def mapping_1(tupled_arg_1: tuple[str, IEncodable]) -> tuple[str, __A_]:
                return (tupled_arg_1[0], tupled_arg_1[1].Encode(helpers_4))

            arg: IEnumerable_1[tuple[str, __A_]] = map_1(mapping_1, values)
            return helpers_4.encode_object(arg)

    return ObjectExpr2559()


def _arrow2566(__unit: None=None) -> Decoder_1[Material]:
    def decode(__unit: None=None) -> Decoder_1[Material]:
        def _arrow2565(get: IGetters) -> Material:
            match_value: str | None
            object_arg: IOptionalGetter = get.Optional
            match_value = object_arg.Field("additionalType", Decode_uri)
            (pattern_matching_result,) = (None,)
            if match_value is None:
                pattern_matching_result = 0

            elif match_value == "Material":
                pattern_matching_result = 0

            else: 
                pattern_matching_result = 1

            if pattern_matching_result == 1:
                object_arg_1: IRequiredGetter = get.Required
                object_arg_1.Field("FailBecauseNotSample", unit)

            def _arrow2560(__unit: None=None) -> str | None:
                object_arg_2: IOptionalGetter = get.Optional
                return object_arg_2.Field("@id", Decode_uri)

            def _arrow2561(__unit: None=None) -> str | None:
                object_arg_3: IOptionalGetter = get.Optional
                return object_arg_3.Field("name", string)

            def _arrow2562(__unit: None=None) -> MaterialType | None:
                object_arg_4: IOptionalGetter = get.Optional
                return object_arg_4.Field("type", ROCrate_decoder_1)

            def _arrow2563(__unit: None=None) -> FSharpList[MaterialAttributeValue] | None:
                arg_11: Decoder_1[FSharpList[MaterialAttributeValue]] = list_1_2(ROCrate_decoder_2)
                object_arg_5: IOptionalGetter = get.Optional
                return object_arg_5.Field("characteristics", arg_11)

            def _arrow2564(__unit: None=None) -> FSharpList[Material] | None:
                arg_13: Decoder_1[FSharpList[Material]] = list_1_2(decode(None))
                object_arg_6: IOptionalGetter = get.Optional
                return object_arg_6.Field("derivesFrom", arg_13)

            return Material(_arrow2560(), _arrow2561(), _arrow2562(), _arrow2563(), _arrow2564())

        return object(_arrow2565)

    return decode(None)


ROCrate_decoder: Decoder_1[Material] = _arrow2566()

def ISAJson_encoder(id_map: Any | None, c: Material) -> IEncodable:
    def f(oa: Material, id_map: Any=id_map, c: Any=c) -> IEncodable:
        def chooser(tupled_arg: tuple[str, IEncodable | None], oa: Any=oa) -> tuple[str, IEncodable] | None:
            def mapping(v_1: IEncodable, tupled_arg: Any=tupled_arg) -> tuple[str, IEncodable]:
                return (tupled_arg[0], v_1)

            return map(mapping, tupled_arg[1])

        def _arrow2570(value: str, oa: Any=oa) -> IEncodable:
            class ObjectExpr2569(IEncodable):
                def Encode(self, helpers: IEncoderHelpers_1[Any]) -> Any:
                    return helpers.encode_string(value)

            return ObjectExpr2569()

        def _arrow2572(value_2: str, oa: Any=oa) -> IEncodable:
            class ObjectExpr2571(IEncodable):
                def Encode(self, helpers_1: IEncoderHelpers_1[Any]) -> Any:
                    return helpers_1.encode_string(value_2)

            return ObjectExpr2571()

        def _arrow2573(oa_1: MaterialAttributeValue, oa: Any=oa) -> IEncodable:
            return ISAJson_encoder_2(id_map, oa_1)

        def _arrow2574(c_1: Material, oa: Any=oa) -> IEncodable:
            return ISAJson_encoder(id_map, c_1)

        values: FSharpList[tuple[str, IEncodable]] = choose(chooser, of_array([try_include("@id", _arrow2570, ROCrate_genID(oa)), try_include("name", _arrow2572, oa.Name), try_include("type", ISAJson_encoder_1, oa.MaterialType), try_include_list_opt("characteristics", _arrow2573, oa.Characteristics), try_include_list_opt("derivesFrom", _arrow2574, oa.DerivesFrom)]))
        class ObjectExpr2575(IEncodable):
            def Encode(self, helpers_2: IEncoderHelpers_1[Any], oa: Any=oa) -> Any:
                def mapping_1(tupled_arg_1: tuple[str, IEncodable]) -> tuple[str, __A_]:
                    return (tupled_arg_1[0], tupled_arg_1[1].Encode(helpers_2))

                arg: IEnumerable_1[tuple[str, __A_]] = map_1(mapping_1, values)
                return helpers_2.encode_object(arg)

        return ObjectExpr2575()

    if id_map is not None:
        def _arrow2576(m_1: Material, id_map: Any=id_map, c: Any=c) -> str:
            return ROCrate_genID(m_1)

        return encode(_arrow2576, f, c, id_map)

    else: 
        return f(c)



ISAJson_allowedFields: FSharpList[str] = of_array(["@id", "@type", "name", "type", "characteristics", "derivesFrom", "@context"])

def _arrow2583(__unit: None=None) -> Decoder_1[Material]:
    def decode(__unit: None=None) -> Decoder_1[Material]:
        def _arrow2582(get: IGetters) -> Material:
            def _arrow2577(__unit: None=None) -> str | None:
                object_arg: IOptionalGetter = get.Optional
                return object_arg.Field("@id", Decode_uri)

            def _arrow2578(__unit: None=None) -> str | None:
                object_arg_1: IOptionalGetter = get.Optional
                return object_arg_1.Field("name", string)

            def _arrow2579(__unit: None=None) -> MaterialType | None:
                object_arg_2: IOptionalGetter = get.Optional
                return object_arg_2.Field("type", ISAJson_decoder_1)

            def _arrow2580(__unit: None=None) -> FSharpList[MaterialAttributeValue] | None:
                arg_7: Decoder_1[FSharpList[MaterialAttributeValue]] = list_1_2(ISAJson_decoder_2)
                object_arg_3: IOptionalGetter = get.Optional
                return object_arg_3.Field("characteristics", arg_7)

            def _arrow2581(__unit: None=None) -> FSharpList[Material] | None:
                arg_9: Decoder_1[FSharpList[Material]] = list_1_2(decode(None))
                object_arg_4: IOptionalGetter = get.Optional
                return object_arg_4.Field("derivesFrom", arg_9)

            return Material(_arrow2577(), _arrow2578(), _arrow2579(), _arrow2580(), _arrow2581())

        return Decode_objectNoAdditionalProperties(ISAJson_allowedFields, _arrow2582)

    return decode(None)


ISAJson_decoder: Decoder_1[Material] = _arrow2583()

__all__ = ["ROCrate_genID", "ROCrate_encoder", "ROCrate_decoder", "ISAJson_encoder", "ISAJson_allowedFields", "ISAJson_decoder"]




from __future__ import annotations
from abc import abstractmethod
from datetime import datetime
from typing import (Any, Protocol)
from ...fable_library.date import (to_universal_time, to_string)
from ...fable_library.list import of_array
from ...thoth_json_core.decode import (map, datetime_local, one_of, bool_1, int_1, float_1, string)
from ...thoth_json_core.types import (Decoder_1, IEncodable, IEncoderHelpers_1)
from ..Cells.fs_cell import DataType

class DateTimeStatic(Protocol):
    @abstractmethod
    def from_time_stamp(self, timestamp: float) -> Any:
        ...


def PyTime_toUniversalTimePy(dt: Any) -> Any:
    timestamp: float = to_universal_time(dt).timestamp()
    return datetime.fromtimestamp(timestamp=timestamp)


Decode_datetime: Decoder_1[Any] = map(PyTime_toUniversalTimePy, datetime_local)

def encode(value: Any=None) -> IEncodable:
    if str(type(value)) == "<class \'str\'>":
        class ObjectExpr323(IEncodable):
            def Encode(self, helpers: IEncoderHelpers_1[Any], value: Any=value) -> Any:
                return helpers.encode_string(value)

        return ObjectExpr323()

    elif str(type(value)) == "<class \'float\'>":
        class ObjectExpr324(IEncodable):
            def Encode(self, helpers_1: IEncoderHelpers_1[Any], value: Any=value) -> Any:
                return helpers_1.encode_decimal_number(value)

        return ObjectExpr324()

    elif str(type(value)) == "<class \'int\'>":
        class ObjectExpr325(IEncodable):
            def Encode(self, helpers_2: IEncoderHelpers_1[Any], value: Any=value) -> Any:
                return helpers_2.encode_signed_integral_number(value)

        return ObjectExpr325()

    elif str(type(value)) == "<class \'bool\'>":
        class ObjectExpr326(IEncodable):
            def Encode(self, helpers_3: IEncoderHelpers_1[Any], value: Any=value) -> Any:
                return helpers_3.encode_bool(value)

        return ObjectExpr326()

    elif isinstance(value, datetime):
        value_5: str = to_string(value, "O", {}).split("+")[0]
        class ObjectExpr327(IEncodable):
            def Encode(self, helpers_4: IEncoderHelpers_1[Any], value: Any=value) -> Any:
                return helpers_4.encode_string(value_5)

        return ObjectExpr327()

    else: 
        class ObjectExpr328(IEncodable):
            def Encode(self, helpers_5: IEncoderHelpers_1[Any], value: Any=value) -> Any:
                return helpers_5.encode_null()

        return ObjectExpr328()



def ctor(b: bool) -> tuple[Any, DataType]:
    return (b, DataType(1))


def ctor_1(i: int) -> tuple[Any, DataType]:
    return (i, DataType(2))


def ctor_2(f: float) -> tuple[Any, DataType]:
    return (f, DataType(2))


def ctor_3(d_3: Any) -> tuple[Any, DataType]:
    return (d_3, DataType(3))


def ctor_4(s: str) -> tuple[Any, DataType]:
    return (s, DataType(0))


decode: Decoder_1[tuple[Any, DataType]] = one_of(of_array([map(ctor, bool_1), map(ctor_1, int_1), map(ctor_2, float_1), map(ctor_3, Decode_datetime), map(ctor_4, string)]))

__all__ = ["PyTime_toUniversalTimePy", "Decode_datetime", "encode", "decode"]




from __future__ import annotations
from ...fable_library.option import default_arg
from ...fable_library.reflection import (TypeInfo, class_type)
from ...fable_library.types import Array
from ..ldobject import (LDNode, LDNode_reflection)

def _expr1541() -> TypeInfo:
    return class_type("ARCtrl.ROCrate.Dataset", None, Dataset, LDNode_reflection())


class Dataset(LDNode):
    def __init__(self, id: str, additional_type: Array[str] | None=None) -> None:
        super().__init__(id, ["schema.org/Dataset"], default_arg(additional_type, []))
        self.init_00408: int = 1


Dataset_reflection = _expr1541

def Dataset__ctor_Z7E7CD246(id: str, additional_type: Array[str] | None=None) -> Dataset:
    return Dataset(id, additional_type)


__all__ = ["Dataset_reflection"]

